"""
PyEQSP utilities module.

Copyright 2026 Paul Leopardi
"""

from math import pi

import numpy as np
from scipy.optimize import newton
from scipy.special import betainc, gamma  # pylint: disable=no-name-in-module

# Tolerance for comparisons close to zero.
_TOLERANCE = float(np.finfo(np.float32).eps)


def asfloat(x):
    """
    a = asfloat(x)

    Convert from a Numpy array to a float when this makes sense.

    It checks if the input is a 0-dimensional array (scalar) or a 1-element array,
    and if so, converts it to a standard Python `float`. Otherwise, it returns
    the input as a NumPy array. This ensures that functions return native Python
    scalars when appropriate (e.g., area of a single region) while still supporting
    vectorized operations.

    Parameters
    ----------
    x : array_like

    Returns
    -------
    a : ndarray or float
        If x is a (), or (1,) or (1,1) array, then a is returned as a float.
        Otherwise a is just x as a Numpy array.

    Examples
    --------
    >>> import numpy as np
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> x0 = 12.789
    >>> a0 = asfloat(x0)
    >>> print(a0)
    12.789
    >>> x1 = [[22.546]]
    >>> a1 = asfloat(x1)
    >>> print(a1)
    22.546
    >>> x2 = [12.789, 22.546]
    >>> a2 = asfloat(x2)
    >>> print(a2)
    [12.789 22.546]
    >>> x3 = np.array([12.789, 22.546])
    >>> a3 = asfloat(x3)
    >>> print(a3)
    [12.789 22.546]
    """
    a = np.asarray(x, dtype=float)
    match a.shape:
        case ():
            return float(a)
        case (1,):
            return float(a[0])
        case (1, 1):
            return float(a[0, 0])
        case _:
            return a


def cart2polar2(x):
    """
    s = cart2polar2(x)

    Convert from Cartesian to spherical coordinates on sphere S^2.

    Parameters
    ----------
    x : ndarray
        An array of real numbers of shape (3, N), where N is any positive integer.
        Each column represents a point in 3D Cartesian coordinates.

    Returns
    -------
    s : ndarray
        An array of shape (2, N), where for each point:
        - s[0, :] is the longitude phi in [0, 2*pi),
        - s[1, :] is the colatitude theta in [0, pi].

    See Also
    --------
    polar2cart

    Notes
    -----
    This function projects any X in R^3 onto the unit sphere S^2 via a line
    through the origin [0, 0, 0]'.
    If X includes the origin, this results in a ValueError exception.

    Examples
    --------
    >>> import numpy as np
    >>> x0 = np.array([[ 0., 0.,  0.,  0.],
    ...                [ 0., 1., -1.,  0.],
    ...                [ 1., 0.,  0., -1.]])
    >>> s0 = cart2polar2(x0)
    >>> print(s0)
    [[0.     1.5708 4.7124 0.    ]
     [0.     1.5708 1.5708 3.1416]]
    >>> x1 = np.array([[ 0., 0.,  0.,  0.],
    ...                [ 0., 1., -1.,  0.],
    ...                [ 0., 0.,  0.,  0.],
    ...                [ 1., 0.,  0., -1.]])
    >>> s1 = cart2polar2(x1)
    Traceback (most recent call last):
        ...
    ValueError: Input x must have shape (3, N)
    >>> x2 = np.array([[ 0., 0.,  0.,  0.],
    ...                [ 0., 0., -1.,  0.],
    ...                [ 1., 0.,  0., -1.]])
    >>> s2 = cart2polar2(x2)
    Traceback (most recent call last):
        ...
    ValueError: Input x must not contain the origin
    """
    x = np.asarray(x)
    if x.shape[0] != 3:
        raise ValueError("Input x must have shape (3, N)")

    # Project any x onto the unit sphere S^2 by normalizing (except for origin)
    norms = np.linalg.norm(x, axis=0)
    # If one or more points is the origin, raise ValueError
    if np.min(norms) < _TOLERANCE:
        raise ValueError("Input x must not contain the origin")
    x_proj = x / norms

    # Spherical coordinates: phi = atan2(y, x), theta = arccos(z)
    phi = np.arctan2(x_proj[1, :], x_proj[0, :]) % (2 * np.pi)
    theta = np.arccos(x_proj[2, :])

    s = np.vstack((phi, theta))
    return s


def polar2cart(s):
    """
    x = polar2cart(s)

    Convert spherical polar to Cartesian coordinates.

    Parameters
    ----------
    s : numpy.ndarray
        Array of real numbers of shape (dim, N) representing N points of S^dim
        in spherical polar coordinates, where dim and N are positive integers.

    Returns
    -------
    x : numpy.ndarray
        Array of shape (dim+1, N) containing the Cartesian coordinates of the
        points represented by the spherical polar coordinates s.

    See Also
    --------
    cart2polar2

    Examples
    --------
    >>> import numpy as np
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> s = np.array([
    ...     [0, pi/2, 3*pi/2, 0],
    ...     [0, pi/2, pi/2,   pi]])
    >>> x = polar2cart(s)
    >>> print(x)
    [[ 0.  0. -0.  0.]
     [ 0.  1. -1.  0.]
     [ 1.  0.  0. -1.]]
    """
    s = np.asarray(s)
    if s.ndim == 1:
        s = s[:, np.newaxis]
        was_1d = True
    else:
        was_1d = False
    dim, n = s.shape
    x = np.zeros((dim + 1, n))
    sinprod = np.ones(n)
    for k in range(dim, 1, -1):
        x[k, :] = sinprod * np.cos(s[k - 1, :])
        sinprod = sinprod * np.sin(s[k - 1, :])
    x[1, :] = sinprod * np.sin(s[0, :])
    x[0, :] = sinprod * np.cos(s[0, :])
    if was_1d:
        return x.flatten()
    return x


def euc2sph_dist(e):
    """
    s = euc2sph_dist(e)

    Convert Euclidean to spherical distance.

    Parameters
    ----------
    e : float or array-like
        A real number or array of real numbers, with ``abs(e) <= 2``.

    Returns
    -------
    s : float or ndarray
        The spherical distance(s), same shape as e.

    Notes
    -----
    The argument e is assumed to satisfy abs(e) <= 2.
    The formula is valid for the unit sphere in all dimensions.

    Examples
    --------
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> print(f"{euc2sph_dist(2):.4f}")
    3.1416
    >>> euc2sph_dist(np.array([0, np.sqrt(2), 2.0]))
    array([0.    , 1.5708, 3.1416])
    >>> print(f"{euc2sph_dist(-2):.4f}")
    -3.1416
    """
    e = np.asarray(e)
    s = 2.0 * np.arcsin(e / 2.0)
    return asfloat(s)


def sph2euc_dist(s):
    """
    e = sph2euc_dist(s)

    Convert spherical distance to Euclidean distance on the unit sphere.

    Parameters
    ----------
    s : float or array_like
        Spherical distance(s), in radians.

    Returns
    -------
    e : float or ndarray
        Euclidean (chord) distance(s), same shape as input.

    Examples
    --------
    >>> from math import pi
    >>> import numpy as np
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> sph2euc_dist(0)
    0.0
    >>> sph2euc_dist(pi)
    2.0
    >>> print(sph2euc_dist(np.array([0, pi/4, pi/2, 3*pi/4, pi])))
    [0.     0.7654 1.4142 1.8478 2.    ]
    >>> print(f"{sph2euc_dist(-pi/2):.4f}")
    -1.4142
    """
    s = np.asarray(s)
    e = 2.0 * np.sin(s / 2.0)
    return asfloat(e)


def euclidean_dist(x, y):
    """
    d = euclidean_dist(x, y)

    Euclidean distance between two points in Cartesian coordinates.

    Parameters
    ----------
    x : array_like, shape (M, N)
        Array of shape (M, N), where each column is a Cartesian vector.
    y : array_like, shape (M, N)
        Array of shape (M, N), where each column is a Cartesian vector.
        The shapes of x and y must be identical.

    Returns
    -------
    d : ndarray, shape (N,)
        The Euclidean distance between corresponding pairs of points in x and y.

    See Also
    --------
    spherical_dist
    euc2sph_dist
    sph2euc_dist


    Examples
    --------
    >>> x = np.array([[0,     0,      0,     0],
    ...               [0,     1,     -1,     0],
    ...               [1,     0,      0,    -1]])
    >>> y = np.array([[ 0,    0,      0,     0],
    ...               [-0.5,  0.866, -0.866, 0.5],
    ...               [ 0.866,0.5,   -0.5,  -0.866]])
    >>> euclidean_dist(x, y)
    array([0.5176, 0.5176, 0.5176, 0.5176])


    """

    x = np.asarray(x)
    y = np.asarray(y)
    if x.shape != y.shape:
        raise ValueError("Input arrays x and y must have the same shape.")

    # Compute the Euclidean distance
    return asfloat(np.sqrt(np.sum((x - y) ** 2, axis=0)))


def spherical_dist(x, y):
    """
    d = spherical_dist(x, y)

    Returns the spherical distance between two arrays of points x and y.

    Parameters
    ----------
    x : array_like
        Array of shape (M, N), where each column is a Cartesian vector.
    y : array_like
        Array of shape (M, N), where each column is a Cartesian vector.
        The shapes of x and y must be identical.

    Returns
    -------
    d : ndarray
        Array of shape (N,), containing spherical distances (in radians)
        between corresponding pairs.

    See Also
    --------
    euclidean_dist
    euc2sph_dist
    sph2euc_dist

    Examples
    --------
    >>> x0 = np.array([[0,     0,      0,     0],
    ...                [0,     1,     -1,     0],
    ...                [1,     0,      0,    -1]])
    >>> y0 = np.array([[ 0,    0,      0,     0],
    ...                [-0.5,  0.866, -0.866, 0.5],
    ...                [ 0.866,0.5,   -0.5,  -0.866]])
    >>> print(spherical_dist(x0, y0))
    [0.5236 0.5236 0.5236 0.5236]
    """
    x = np.asarray(x, dtype=float)
    y = np.asarray(y, dtype=float)
    if x.shape != y.shape:
        raise ValueError("x and y must both have shape (M, N)")

    # Compute dot product for each point (across columns)
    dots = np.sum(x * y, axis=0)
    # Clip dot product to [-1, 1] to avoid numerical errors outside acos domain
    dots = np.clip(dots, -1.0, 1.0)
    return asfloat(np.arccos(dots))


def area_of_sphere(dim):
    """
    a = area_of_sphere(dim)

    Returns the area of the unit sphere S^dim.

    Parameters
    ----------
    dim : int or array-like of int
        Dimension(s) of the sphere(s). Must be positive integer(s).

    Returns
    -------
    a : float or np.ndarray
        Area(s) of the unit sphere(s).

    Notes
    -----
    The area of S^dim is defined via the Lebesgue measure on S^dim
    inherited from its embedding in R^(dim+1).

    The area is calculated as:
        a = 2 * pi^((dim+1)/2) / gamma((dim+1)/2)

    References
    ----------
    [Mue98] p39.

    See Also
    --------
    volume_of_ball


    Examples
    --------
    >>> area_of_sphere(range(1, 8))
    array([ 6.2832, 12.5664, 19.7392, 26.3189, 31.0063, 33.0734, 32.4697])
    """
    dim = np.asarray(dim)
    power = (dim + 1) / 2
    area = np.asarray(2.0 * pi**power / gamma(power))
    return asfloat(area)


def volume_of_ball(dim):
    """
    v = volume_of_ball(dim)

    Volume of the unit ball B^dim in R^dim.

    Parameters
    ----------
    dim : int or array-like
        Dimension(s) of the ball(s). Must be positive integer(s).

    Returns
    -------
    v : float or ndarray
        Volume(s) of the unit ball(s).

    Notes
    -----
    The volume of B^dim is defined via the Lebesgue measure on R^dim.

    References
    ----------
    [WeiMW].

    See Also
    --------
    area_of_sphere

    Examples
    --------
    >>> import numpy as np
    >>> volume_of_ball(range(1, 8))
    array([2.    , 3.1416, 4.1888, 4.9348, 5.2638, 5.1677, 4.7248])
    """
    dim = np.asarray(dim)
    return asfloat(area_of_sphere(dim - 1) / dim)


def area_of_ideal_region(dim, N):
    """
    a = area_of_ideal_region(dim, N)

    Area of one region of an EQ partition.

    This function returns the area of one of N equal-area regions of
    a unit sphere S^dim, i.e., 1/N times area_of_sphere(dim).

    Parameters
    ----------
    dim : int
        Dimension of the sphere (must be positive).
    N : int or array-like of int
        Number(s) of regions (must be positive).

    Returns
    -------
    a : float or numpy.ndarray with the same shape as N
        Area(s) of the ideal region(s).

    See Also
    --------
    area_of_sphere

    Examples
    --------
    >>> area_of_ideal_region(3, range(1, 7))
    array([19.7392,  9.8696,  6.5797,  4.9348,  3.9478,  3.2899])
    """
    area = area_of_sphere(dim) / np.array(N)
    return asfloat(area)


def ideal_collar_angle(dim, N):
    """
    The ideal angle for spherical collars of an EQ partition.

    Parameters
    ----------
    dim : int
        Dimension of sphere.
    N : int or array-like
        Number of regions.

    Returns
    -------
    angle : float or np.ndarray
        The ideal angle(s).

    Notes
    -----
    The ideal collar angle is determined by the side of a dim-dimensional
    hypercube of the same volume as the area of single region of S^dim.

    See Also
    --------
    area_of_ideal_region

    Examples
    --------
    >>> print(f"{ideal_collar_angle(2, 10):.4g}")
    1.121
    >>> np.round(ideal_collar_angle(3, np.arange(1,7)), 4)
    array([2.7026, 2.145 , 1.8739, 1.7025, 1.5805, 1.4873])
    """
    return asfloat(area_of_ideal_region(dim, N) ** (1 / dim))


def area_of_cap(dim, s_cap):
    """
    a = area_of_cap(dim, s_cap)

    Area of spherical cap on S^dim of spherical radius s_cap.

    Parameters
    ----------
    dim : int
        Positive integer, the dimension of the sphere.
    s_cap : float or array_like
        Spherical radius/radii of the cap(s), in [0, pi].

    Returns
    -------
    area : float or ndarray
        Area(s) of the spherical cap(s).

    Notes
    -----
    The area is defined via the Lebesgue measure on S^dim inherited from
    its embedding in R^(dim+1).

    For dim <= 2, and for dim==3 (when pi/6 <= s_cap <= 5*pi/6),
    AREA is calculated in closed form, using the analytic solution of
    the definite integral.

    Otherwise, AREA is calculated using the incomplete Beta function ratio.

    References
    ----------
    [LeGS01 Lemma 4.1 p255]

    See Also
    --------
    sradius_of_cap

    Examples
    --------
    >>> print(f"{area_of_cap(2, pi/2):.4f}")
    6.2832
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> print(area_of_cap(3, np.linspace(0, pi, 5)))
    [ 0.      1.7932  9.8696 17.946  19.7392]
    """
    s_cap = np.asarray(s_cap)
    match dim:
        case 1:
            area = 2.0 * s_cap
        case 2:
            area = 4.0 * pi * np.sin(s_cap / 2.0) ** 2
        case 3:
            shape = s_cap.shape
            s_cap_flat = s_cap.ravel()
            area = np.zeros_like(s_cap_flat, dtype=np.float64)
            MIN_TROPICAL = pi / 6
            MAX_TROPICAL = 5 * pi / 6
            near_pole = (s_cap_flat < MIN_TROPICAL) | (s_cap_flat > MAX_TROPICAL)
            # Use incomplete beta function ratio near poles
            area[near_pole] = area_of_sphere(dim) * betainc(
                dim / 2, dim / 2, np.sin(s_cap_flat[near_pole] / 2) ** 2
            )
            # Use closed form in the tropics
            s_cap_trop = s_cap_flat[~near_pole]
            area[~near_pole] = (2 * s_cap_trop - np.sin(2 * s_cap_trop)) * pi
            area = area.reshape(shape)
        case _:
            area = area_of_sphere(dim) * betainc(
                dim / 2, dim / 2, np.sin(s_cap / 2) ** 2
            )
    return asfloat(area)


def sradius_of_cap(dim, area):
    """
    s_cap = sradius_of_cap(dim, area)

    Spherical radius of a spherical cap of given area on S^dim.

    Parameters
    ----------
    dim : int
        Dimension of the sphere (must be >= 1).
    area : float or array-like
        Area(s) of the cap(s).

    Returns
    -------
    s_cap : float or ndarray
        Spherical radius/radii in [0, pi], same shape as area.

    Notes
    -----
    For dim <= 2, the result is exact (closed form).
    For dim > 2, the result is found numerically.

    See Also
    --------
    area_of_cap

    Examples
    --------

    >>> import numpy as np
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> area = area_of_sphere(2) / 2
    >>> print(f"{sradius_of_cap(2, area):.4f}")
    1.5708

    >>> areas = np.linspace(0, 4, 5) * area_of_sphere(3) / 4
    >>> sradius_of_cap(3, areas)
    array([0.    , 1.1549, 1.5708, 1.9867, 3.1416])
    """
    area = np.asarray(area)
    sphere_area = area_of_sphere(dim)

    if np.any(area < 0):
        raise ValueError("Area must be non-negative.")
    if np.any(area > sphere_area * (1 + 1e-10)):
        raise ValueError(f"Area {np.max(area)} exceeds area of sphere {sphere_area}.")

    if dim == 1:
        s_cap = np.clip(area / 2, 0, pi)
    elif dim == 2:
        # Avoid nan for area slightly > sphere_area by clipping to [0, 1]
        arg = np.clip(area / sphere_area, 0.0, 1.0)
        s_cap = 2 * np.arcsin(np.sqrt(arg))
    else:
        orig_shape = area.shape
        flat_area = np.ravel(area)
        s_cap = np.zeros_like(flat_area, dtype=float)

        # Handle cases matching or exceeding full sphere area
        full_idx = flat_area >= sphere_area
        s_cap[full_idx] = pi

        # Handle zero or negative area (ValueError caught negative areas)
        zero_idx = flat_area <= 0
        s_cap[zero_idx] = 0.0

        # Process remaining cases
        calc_idx = ~(full_idx | zero_idx)
        if np.any(calc_idx):
            ak_calc = flat_area[calc_idx]

            # Mirror areas greater than half the sphere for better convergence
            flipped = 2 * ak_calc > sphere_area
            ak_target = np.where(flipped, sphere_area - ak_calc, ak_calc)

            # Pre-allocate inputs for vectorized optimization
            # Start with a good initial guess (e.g. midpoint of remaining space)
            x0 = np.full_like(ak_target, pi / 2)

            def area_diff(s):
                # Vectorized difference function for Newton solver.
                return area_of_cap(dim, s) - ak_target

            def area_diff_prime(s):
                # Derivative of area of cap with respect to s.
                # Strictly: d/ds Area(dim, s) = Area(dim-1, s) based on S^{dim-1}
                # radius sin(s).
                return area_of_sphere(dim - 1) * (np.sin(s) ** (dim - 1))

            # Find roots using vectorized newton
            sk = newton(area_diff, x0, fprime=area_diff_prime)

            # Ensure roots are clamped within [0, pi]
            sk = np.clip(sk, 0.0, pi)

            # Map back flipped variants
            s_cap[calc_idx] = np.where(flipped, pi - sk, sk)

        s_cap = s_cap.reshape(orig_shape)
    return asfloat(s_cap)


def area_of_collar(dim, a_top, a_bot):
    """
    a = area_of_collar(dim, a_top, a_bot)

    Area of a spherical collar.

    Parameters
    ----------
    dim : int
        Positive integer, the dimension of the sphere.
    a_top : float or array-like
        Top (smaller) spherical radius/radii, in [0, pi].
    a_bot : float or array-like
        Bottom (larger) spherical radius/radii, in [0, pi].

    Returns
    -------
    a : float or ndarray
        Area(s) of the spherical collar(s).

    Notes
    -----
    a_top and a_bot must have the same shape.
    The area is defined via the Lebesgue measure on S^dim
    inherited from its embedding in R^(dim+1).

    References
    ----------
    [LeGS01 Lemma 4.1 p255]

    See Also
    --------
    area_of_cap

    Examples
    --------
    >>> import numpy as np
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> area_of_collar(2, np.array([0, 1, 2]), np.array([1, 2, 3]))
    array([2.8884, 6.0095, 3.6056])
    """
    a_top = np.asarray(a_top)
    a_bot = np.asarray(a_bot)
    return asfloat(area_of_cap(dim, a_bot) - area_of_cap(dim, a_top))


def x2stereo(x):
    """
    Stereographic projection of Euclidean points.

    Parameters
    ----------
    x : ndarray
        Points in R^{dim+1}, shape (dim+1, N).

    Returns
    -------
    result : ndarray
        Projected points in R^dim, shape (dim, N).
    """
    x = np.asarray(x)
    dim = x.shape[0] - 1

    last = x[dim, :]
    mask = np.isclose(last, 1.0)

    scale = np.ones(x.shape[1])
    scale[~mask] = 1.0 - last[~mask]

    with np.errstate(divide="ignore"):
        result = x[:dim, :] / scale

    result[:, mask] = np.nan
    return result


def x2eqarea(x):
    """
    Equal area projection of Euclidean points.

    Parameters
    ----------
    x : ndarray
        Points in R^{dim+1}, shape (dim+1, N).

    Returns
    -------
    result : ndarray
        Projected points in R^dim, shape (dim, N).
    """
    x = np.asarray(x)
    dim = x.shape[0] - 1
    last = x[dim, :]

    theta = np.arccos(np.clip(-last, -1.0, 1.0))
    a_cap = area_of_cap(dim, theta)
    v_ball = volume_of_ball(dim)
    r = (a_cap / v_ball) ** (1.0 / dim)

    sin_theta = np.sin(theta)
    mask = np.isclose(sin_theta, 0.0)

    scale = np.zeros_like(theta)
    scale[~mask] = r[~mask] / sin_theta[~mask]

    result = np.zeros((dim, x.shape[1]))
    result[:, ~mask] = x[:dim, ~mask] * scale[~mask]
    return result


if __name__ == "__main__":
    import doctest  # pragma: no cover

    # pragma: no cover
    doctest.testmod()  # pragma: no cover



from math import cos, gcd, pi, sin

import numpy as np

from ..utilities import (
    area_of_collar,
    area_of_ideal_region,
    sradius_of_cap,
)


def bot_cap_region(dim, a_cap):
    """
    South polar (bottom) cap region of EQ partition.

    Parameters
    ----------
    dim : int
        Dimension of the sphere S^dim.
    a_cap : float
        Spherical radius of the bottom cap.

    Returns
    -------
    region : ndarray
        (dim, 2) array of two points in spherical polar coordinates
        representing the South polar cap region.

    See Also
    --------
    top_cap_region, sphere_region

    Raises
    ------
    ValueError
        If inputs are not appropriate numeric values.

    Notes
    -----
    The returned array shape is (dim, 2), columns represent bounds.

    Examples
    --------
    >>> np.set_printoptions(precision=4)
    >>> bot_cap_region(1, pi/6)
    array([[5.7596, 6.2832]])
    >>> bot_cap_region(2, pi/6)
    array([[0.    , 6.2832],
           [2.618 , 3.1416]])
    >>> bot_cap_region(3, pi/6)
    array([[0.    , 6.2832],
           [0.    , 3.1416],
           [2.618 , 3.1416]])
    """
    if dim == 1:
        return np.array([[2 * pi - a_cap, 2 * pi]])
    sphere_region_1 = sphere_region(dim - 1)
    first_col = np.append(sphere_region_1[:, 0], pi - a_cap)
    second_col = np.append(sphere_region_1[:, 1], pi)
    return np.column_stack([first_col, second_col])


def cap_colats(dim, N, c_polar, n_regions):
    """
    Colatitudes of spherical caps enclosing cumulative sum of regions.

    Parameters
    ----------
    dim : int
        Dimension of the sphere S^dim.
    N : int
        Number of regions in the partition.
    c_polar : float
        Colatitude of the North polar cap.
    n_regions : array_like
        List of region counts per collar and cap.

    Returns
    -------
    c_caps : ndarray
        Increasing array of colatitudes enclosing cumulative region sums.

    See Also
    --------
    polar_colat, sradius_of_cap, area_of_ideal_region

    Raises
    ------
    ValueError
        If inputs are inconsistent in length or type.

    Notes
    -----
    Length equals n_collars + 2. c_caps[0] = c_polar, last = pi.

    Examples
    --------
    >>> np.set_printoptions(precision=4)
    >>> dim = 2
    >>> N = 4
    >>> c_polar = polar_colat(dim, N)
    >>> n_regions = np.array([1, 2, 1])
    >>> cap_colats(dim, N, c_polar, n_regions)
    array([1.0472, 2.0944, 3.1416])
    """
    n_regions = np.asarray(n_regions)
    c_caps = np.zeros_like(n_regions, dtype=float)
    c_caps[0] = c_polar
    ideal_region_area = area_of_ideal_region(dim, N)
    n_collars = n_regions.size - 2
    subtotal_n_regions = 1
    for collar_n in range(1, n_collars + 1):
        subtotal_n_regions = subtotal_n_regions + n_regions[collar_n]
        c_caps[collar_n] = sradius_of_cap(dim, subtotal_n_regions * ideal_region_area)
    c_caps[1 + n_collars] = pi
    return c_caps


def centres_of_regions(regions):
    """
    Centre points of given regions.

    Parameters
    ----------
    regions : ndarray
        (dim, 2, N) array of region bounds in spherical polar coordinates.

    Returns
    -------
    points : ndarray
        (dim, N) array of region centre points.

    See Also
    --------
    sphere_region

    Raises
    ------
    AssertionError
        If input dimensions are inconsistent.

    Notes
    -----
    Uses floating point tolerance for equality. Azimuth mod 2*pi, others mod pi.

    Examples
    --------
    >>> regions = np.array([[[0, 1], [2, 3]], [[0, 1], [2, 3]]])
    >>> centres_of_regions(regions)
    array([[1., 2.],
           [0., 2.]])
    """
    tol = np.finfo(float).eps * 2**5
    regions = np.asarray(regions, dtype=float)
    regions = regions.copy()
    if regions.ndim == 2:
        regions = regions[:, :, np.newaxis]
    dim = regions.shape[0]
    N = regions.shape[2]
    points = np.zeros((dim, N), dtype=float)
    top = regions[:, 0, :]
    bot = regions[:, 1, :]
    zero_bot = np.abs(bot[0, :]) < tol
    bot[0, zero_bot] = 2 * pi
    equal_bot = np.abs(bot[0, :] - top[0, :]) < tol
    bot[0, equal_bot] = top[0, equal_bot] + 2 * pi
    twopi_bot = np.abs(bot[0, :] - top[0, :] - 2 * pi) < tol
    points[0, twopi_bot] = 0.0
    mask_other = ~twopi_bot
    points[0, mask_other] = np.mod(
        (bot[0, mask_other] + top[0, mask_other]) / 2.0, 2 * pi
    )
    for k in range(1, dim):
        pi_bot = np.abs(bot[k, :] - pi) < tol
        points[k, pi_bot] = pi
        zero_top = np.abs(top[k, :]) < tol
        points[k, zero_top] = 0.0
        all_else = ~(zero_top | pi_bot)
        points[k, all_else] = np.mod((top[k, all_else] + bot[k, all_else]) / 2.0, pi)
    return points


def circle_offset(n_top, n_bot, extra_twist=False):
    """
    Maximize minimum distance of centre points for S^2 collars.

    Parameters
    ----------
    n_top : int
        Number of points in top circle.
    n_bot : int
        Number of points in bottom circle.
    extra_twist : bool, optional
        If True, add an extra sector twist.

    Returns
    -------
    offset : float
        Offset in multiples of rotations.

    See Also
    --------
    rot3, s2_offset

    Notes
    -----
    Offset includes a half-sector twist, max-min separation, and optional twist.

    Examples
    --------
    >>> def disp(x):
    ...     print(f'{x:.5g}')
    >>> disp(circle_offset(3, 4))
    6.9389e-18
    >>> disp(circle_offset(3, 4, extra_twist=True))
    1.5
    >>> disp(circle_offset(7, 11))
    -0.019481
    """
    if n_top == 0 or n_bot == 0:
        raise ValueError("n_top and n_bot must be positive integers")
    offset = (1.0 / n_bot - 1.0 / n_top) / 2.0 + gcd(int(n_top), int(n_bot)) / (
        2.0 * n_top * n_bot
    )
    if extra_twist:
        twist = 6
        offset = offset + twist / float(n_bot)
    return offset


def ideal_region_list(dim, N, c_polar, n_collars):
    """
    Ideal real number of regions in each zone.

    Parameters
    ----------
    dim : int
        Dimension of sphere.
    N : int
        Number of regions.
    c_polar : float
        North polar colatitude.
    n_collars : int
        Number of collars.

    Returns
    -------
    r_regions : ndarray
        Array of ideal real region counts, length n_collars+2.

    See Also
    --------
    ideal_collar_angle, num_collars, polar_colat

    Notes
    -----
    r_regions[0] and r_regions[-1] are 1. Sum equals N.

    Examples
    --------
    >>> dim = 2
    >>> N = 4
    >>> c_polar = polar_colat(dim, N)
    >>> n_collars = 1
    >>> ideal_region_list(dim, N, c_polar, n_collars)
    array([1., 2., 1.])
    """
    r_regions = np.zeros(2 + int(n_collars), dtype=float)
    r_regions[0] = 1.0
    if n_collars > 0:
        a_fitting = (pi - 2.0 * c_polar) / float(n_collars)
        ideal_region_area = area_of_ideal_region(dim, N)
        for collar_n in range(1, n_collars + 1):
            a1 = c_polar + (collar_n - 1) * a_fitting
            a2 = c_polar + collar_n * a_fitting
            ideal_collar_area = area_of_collar(dim, a1, a2)
            r_regions[collar_n] = ideal_collar_area / ideal_region_area
    r_regions[-1] = 1.0
    return r_regions


def num_collars(N, c_polar, a_ideal):
    """
    Number of collars between polar caps.

    Parameters
    ----------
    N : int or array_like
        Number of regions.
    c_polar : float or array_like
        North polar colatitude.
    a_ideal : float or array_like
        Ideal collar angle.

    Returns
    -------
    n_collars : int or ndarray
        Number of collars.

    See Also
    --------
    ideal_collar_angle, ideal_region_list, polar_colat

    Notes
    -----
    Zero for N <= 2 or a_ideal <= 0.

    Examples
    --------
    >>> dim = 2
    >>> N = 4
    >>> c_polar = polar_colat(dim, N)
    >>> a_ideal = 1.0471975512
    >>> num_collars(N, c_polar, a_ideal)
    1
    """
    N_arr = np.asarray(N)
    c_polar_arr = np.asarray(c_polar)
    a_ideal_arr = np.asarray(a_ideal)
    n_collars = np.zeros_like(N_arr, dtype=int)
    enough = (N_arr > 2) & (a_ideal_arr > 0)
    if np.any(enough):
        val = np.round((pi - 2.0 * c_polar_arr[enough]) / a_ideal_arr[enough])
        val = np.maximum(1, val).astype(int)
        n_collars[enough] = val
    if np.isscalar(N):
        return int(n_collars)
    return n_collars


def polar_colat(dim, N):
    """
    Colatitude of North polar (top) spherical cap.

    Parameters
    ----------
    dim : int
        Dimension of sphere.
    N : int or array_like
        Number of regions.

    Returns
    -------
    c_polar : float or ndarray
        North polar colatitude.

    See Also
    --------
    ideal_collar_angle, ideal_region_list, num_collars

    Notes
    -----
    c_polar = pi for N==1, pi/2 for N==2, else computed from ideal region.

    Examples
    --------
    >>> def disp(x):
    ...     print(f'{x:.5g}')
    >>> disp(polar_colat(2, 4))
    1.0472
    >>> disp(polar_colat(2, 10))
    0.6435
    >>> disp(polar_colat(3, 6))
    0.98448
    """
    if np.isscalar(N):
        if N == 1:
            return pi
        if N == 2:
            return pi / 2.0
        area = area_of_ideal_region(dim, N)
        return sradius_of_cap(dim, area)
    N_arr = np.asarray(N)
    c_polar = np.zeros_like(N_arr, dtype=float)
    c_polar[N_arr == 1] = pi
    c_polar[N_arr == 2] = pi / 2.0
    mask = N_arr > 2
    if np.any(mask):
        areas = area_of_ideal_region(dim, N_arr[mask])
        c_polar[mask] = sradius_of_cap(dim, areas)
    return c_polar


def rot3(axis, angle):
    """
    R^3 rotation about a coordinate axis.

    Parameters
    ----------
    axis : {1,2,3}
        Axis index (1=x, 2=y, 3=z).
    angle : float
        Rotation angle in radians.

    Returns
    -------
    R : ndarray
        3x3 rotation matrix.

    Examples
    --------
    >>> rot3(1, pi/6)
    array([[ 1.   ,  0.   ,  0.   ],
           [ 0.   ,  0.866, -0.5  ],
           [ 0.   ,  0.5  ,  0.866]])
    >>> rot3(2, pi/6)
    array([[ 0.866,  0.   , -0.5  ],
           [ 0.   ,  1.   ,  0.   ],
           [ 0.5  ,  0.   ,  0.866]])
    >>> rot3(3, pi/6)
    array([[ 0.866, -0.5  ,  0.   ],
           [ 0.5  ,  0.866,  0.   ],
           [ 0.   ,  0.   ,  1.   ]])
    """
    c = cos(angle)
    s = sin(angle)
    if axis == 1:
        R = np.array([[1.0, 0.0, 0.0], [0.0, c, -s], [0.0, s, c]])
    elif axis == 2:
        R = np.array([[c, 0.0, -s], [0.0, 1.0, 0.0], [s, 0.0, c]])
    elif axis == 3:
        R = np.array([[c, -s, 0.0], [s, c, 0.0], [0.0, 0.0, 1.0]])
    else:
        raise ValueError("axis must be 1, 2, or 3")
    return R


def round_to_naturals(N, r_regions):
    """
    Round off a given list of numbers of regions.

    Parameters
    ----------
    N : int
        Total number of regions.
    r_regions : array_like
        Ideal real region counts (length n_collars+2).

    Returns
    -------
    n_regions : ndarray
        Integer region counts, length as r_regions, sum equals N.

    Raises
    ------
    AssertionError
        If sum does not equal N.

    Notes
    -----
    Uses discrepancy tracking for rounding.

    Examples
    --------
    >>> round_to_naturals(4, [1.0, 2.0, 1.0])
    array([1, 2, 1])
    """
    r_regions = np.asarray(r_regions, dtype=float)
    n_regions = np.zeros_like(r_regions, dtype=int)
    discrepancy = 0.0
    for i in range(r_regions.size):
        n_regions[i] = int(round(r_regions[i] + discrepancy))
        discrepancy = discrepancy + r_regions[i] - float(n_regions[i])
    assert n_regions.sum() == int(N), "Sum of result n_regions does not equal N==%g" % N
    return n_regions


def s2_offset(points_1):
    """
    Experimental offset rotation of S^2.

    Parameters
    ----------
    points_1 : ndarray
        2 x M array representing M points in spherical polar coordinates.

    Returns
    -------
    rotation : ndarray
        3x3 rotation matrix moving north pole to specified location.

    See Also
    --------
    rot3, circle_offset

    Notes
    -----
    For equal colats, average azimuth used; else, azimuth+pi.

    Examples
    --------
    >>> s = np.array([[0., 0.785398, 2.356194, 3.9270, 5.49778, 0.],
    ...               [0., 1.570796, 1.570796, 1.570796, 1.570796, 3.1416]])
    >>> np.set_printoptions(suppress=True, precision=4)
    >>> s2_offset(s)
    array([[ 0.    ,  0.7071,  0.7071],
           [-1.    ,  0.    ,  0.    ],
           [-0.    , -0.7071,  0.7071]])
    """
    points_1 = np.asarray(points_1, dtype=float)
    n_in_collar = points_1.shape[1]
    if n_in_collar > 2:
        if (n_in_collar > 3) and (points_1[1, 1] == points_1[1, 2]):
            a_3 = (points_1[0, 1] + points_1[0, 2]) / 2.0
        else:
            a_3 = points_1[0, 1] + pi  # pragma: no cover
        a_2 = points_1[1, 1] / 2.0
    else:
        a_3 = 0.0  # pragma: no cover
        a_2 = pi / 2.0  # pragma: no cover
    return np.dot(rot3(2, -a_2), rot3(3, -a_3))


def sphere_region(dim):
    """
    Sphere represented as single region of EQ partition.

    Parameters
    ----------
    dim : int
        Dimension of sphere S^dim.

    Returns
    -------
    region : ndarray
        (dim, 2) array representing the whole sphere region.

    See Also
    --------
    bot_cap_region, top_cap_region

    Examples
    --------
    >>> sphere_region(1)
    array([[0.    , 6.2832]])
    >>> sphere_region(2)
    array([[0.    , 6.2832],
           [0.    , 3.1416]])
    >>> sphere_region(3)
    array([[0.    , 6.2832],
           [0.    , 3.1416],
           [0.    , 3.1416]])
    """
    if dim == 1:
        return np.array([[0.0, 2.0 * pi]])
    sphere_region_1 = sphere_region(dim - 1)
    first_col = np.append(sphere_region_1[:, 0], 0.0)
    second_col = np.append(sphere_region_1[:, 1], pi)
    return np.column_stack([first_col, second_col])


def top_cap_region(dim, a_cap):
    """
    North polar (top) cap region of EQ partition.

    Parameters
    ----------
    dim : int
        Dimension of sphere S^dim.
    a_cap : float
        Spherical radius of the top cap.

    Returns
    -------
    region : ndarray
        (dim, 2) array of two points in spherical polar coordinates
        representing the North polar cap region.

    See Also
    --------
    bot_cap_region, sphere_region

    Notes
    -----
    The returned array shape is (dim, 2).

    Examples
    --------
    >>> import math
    >>> top_cap_region(1, pi/6)
    array([[0.    , 0.5236]])
    >>> top_cap_region(2, pi/6)
    array([[0.    , 6.2832],
           [0.    , 0.5236]])

    >>> top_cap_region(3, pi/6)
    array([[0.    , 6.2832],
           [0.    , 3.1416],
           [0.    , 0.5236]])
    """
    if dim == 1:
        return np.array([[0.0, a_cap]])
    sphere_region_1 = sphere_region(dim - 1)
    first_col = np.append(sphere_region_1[:, 0], 0.0)
    second_col = np.append(sphere_region_1[:, 1], a_cap)
    return np.column_stack([first_col, second_col])



"""
PyEQSP Visualizations module.
3D visualizations using Mayavi.

Copyright 2025 Paul Leopardi.
For licensing, see COPYING.
"""

import matplotlib.pyplot as plt
import numpy as np

from .partitions import eq_point_set, eq_regions
from .utilities import (
    polar2cart,
    x2eqarea,
    x2stereo,
)

try:
    from mayavi import mlab
except ImportError as exc:  # pragma: no cover
    raise ImportError(
        "Mayavi is not installed. Please install it with: pip install 'eqsp[mayavi]'"
    ) from exc

PROJ_NAME = {"eqarea": "equal area", "stereo": "stereographic"}


def show_s2_sphere(opacity=0.95, color=(0, 1, 0)):
    """
    Illustrate the unit sphere S^2.
    """
    u, v = np.mgrid[0 : 2 * np.pi : 50j, 0 : np.pi : 50j]
    x = np.cos(u) * np.sin(v)
    y = np.sin(u) * np.sin(v)
    z = np.cos(v)

    mlab.mesh(x, y, z, color=color, opacity=opacity)


def show_r3_point_set(
    points,
    show_sphere=False,
    scale_factor=0.1,
    save_file=None,
    **kwargs,
):
    """
    3D illustration of a point set.
    """
    if show_sphere:
        show_s2_sphere()

    # Mayavi points3d expects x, y, z, s (scalar) - s is optional
    mlab.points3d(
        points[0, :],
        points[1, :],
        points[2, :],
        scale_factor=scale_factor,
        color=(1, 0, 0),
        **kwargs,
    )

    if save_file:
        mlab.savefig(save_file)


def show_s2_region(region, N, fidelity=32):
    """
    Illustrate a region of S^2.
    """
    # pylint: disable=no-member
    tol = np.finfo(float).eps * 32
    dim = region.shape[0]
    t = region[:, 0]
    b = region[:, 1]

    if abs(b[0]) < tol:
        b[0] = 2 * np.pi  # pragma: no cover
    pseudo = abs(t[0]) < tol and abs(b[0] - 2 * np.pi) < tol

    h = np.linspace(0, 1, fidelity)
    r = np.sqrt(1.0 / N) / 12.0

    for k in range(dim):
        if pseudo and k >= 1:
            continue
        j = np.arange(dim)
        j = np.roll(j, -k)

        s_curve = np.zeros((dim, fidelity))
        idx_vary = j[0]
        idx_fixed = j[1:]

        s_curve[idx_vary, :] = t[idx_vary] + (b[idx_vary] - t[idx_vary]) * h
        for i_f in idx_fixed:
            s_curve[i_f, :] = t[i_f]

        x_curve = polar2cart(s_curve)

        # Mayavi plot3d for tube plotting
        mlab.plot3d(x_curve[0], x_curve[1], x_curve[2], tube_radius=r, color=(0, 0, 1))


def show_s2_partition(
    N,
    *,
    extra_offset=False,
    show_points=True,
    show_sphere=True,
    title="long",
    title_pos=(0.2, 0.85),
    show=True,
    save_file=None,
    **_kwargs,
):
    """
    3D illustration of an EQ partition of S^2 into N regions.

    Parameters
    ----------
    N : int
        Number of regions.
    extra_offset : bool, optional
        Use extra offsets. Default False.
    show_points : bool, optional
        Show centre points. Default True.
    show_sphere : bool, optional
        Show unit sphere. Default True.
    title : str, optional
        Title text. Special values: 'long', 'short', 'none'.
        'long' uses a default multi-line description.
        'short' uses 'EQ(2, N)'.
        'none' shows no title.
        Any other string is used as the title text.
    title_pos : tuple, optional
        (x, y) position of the title in figure coordinates (0 to 1).
        Default is (0.2, 0.85).
    **kwargs
        Passed to Mayavi functions.

    Examples
    --------
    >>> from eqsp.visualizations import show_s2_partition
    >>> from mayavi import mlab
    >>> mlab.options.offscreen = True
    >>> try:
    ...     show_s2_partition(4, title='short', show_points=False)
    ...     print("Success") # Crude check as Mayavi is hard to doctest
    ... except ImportError:
    ...     print("Mayavi not installed")
    Success
    """
    title_text = None
    if title == "none":
        show_title = False
    else:
        show_title = True
        if title == "long":
            title_text = (
                f"Recursive zonal equal area partition of S^2\ninto {N} regions."
            )
        elif title == "short":
            title_text = f"EQ(2, {N})"
        else:
            title_text = title

    # Set default figure size if none exists
    if mlab.get_engine().current_scene is None:
        mlab.figure(bgcolor=(1, 1, 1), size=(800, 800))

    if show_sphere:
        show_s2_sphere(opacity=0.95)

    R = eq_regions(2, N, extra_offset)
    for i in range(N - 1, 0, -1):
        show_s2_region(R[:, :, i], N)

    if show_points:
        points = eq_point_set(2, N, extra_offset)
        show_r3_point_set(points, show_sphere=False)

    if show_title:
        # Use mlab.text for precise control over size and position
        mlab.text(title_pos[0], title_pos[1], title_text, width=0.6, color=(0, 0, 0))

    if save_file:
        mlab.savefig(save_file)

    if show:
        mlab.show()


def project_point_set(
    points,
    proj="stereo",
    scale_factor=0.1,
    color=(1, 0, 0),
    show=True,
    save_file=None,
    **kwargs,
):
    """
    Use projection to illustrate a point set of S^2 or S^3.

    Parameters
    ----------
    points : ndarray
        Array of shape (dim+1, N) containing centre points of each region in
        Cartesian coordinates.
    proj : {'stereo', 'eqarea'}, optional
        Projection type. Default 'stereo'.
    scale_factor : float, optional
        Scale factor for points. Default 0.1.
    color : tuple, optional
        Colour of points in RGB format (0 to 1). Default (1, 0, 0).
    **kwargs
        Passed to Mayavi plotting functions.

    Examples
    --------
    >>> from eqsp.visualizations import project_point_set
    >>> import numpy as np
    >>> from mayavi import mlab
    >>> mlab.options.offscreen = True
    >>> points = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]).T
    >>> try:
    ...     project_point_set(points, proj='eqarea')
    ...     print("Success")
    ... except ImportError:
    ...     print("Mayavi not installed")
    Success
    """
    points = np.asarray(points)
    dim = points.shape[0] - 1
    if dim not in [2, 3]:
        raise ValueError("Points must be in R^3 (S^2) or R^4 (S^3)")

    if proj == "stereo":
        projector = x2stereo
    elif proj == "eqarea":
        projector = x2eqarea
    else:
        raise ValueError("proj must be 'stereo' or 'eqarea'")

    # Set default figure size if none exists
    if mlab.get_engine().current_scene is None:
        mlab.figure(bgcolor=(1, 1, 1), size=(800, 800))

    t = projector(points)

    # Mayavi points3d
    # scale_factor and color are explicit arguments now, but kwargs can override?
    # Actually explicit args take precedence in this implementation, assume user passes
    # them.

    if dim == 2:
        # Project to z=0 for S^2 -> R^2
        mlab.points3d(
            t[0, :],
            t[1, :],
            np.zeros_like(t[0, :]),
            scale_factor=scale_factor,
            color=color,
            **kwargs,
        )
    elif dim == 3:
        # S^3 -> R^3
        mlab.points3d(
            t[0, :], t[1, :], t[2, :], scale_factor=scale_factor, color=color, **kwargs
        )

    if save_file:
        mlab.savefig(save_file)

    if show:
        mlab.show()


def project_s3_partition(
    N,
    *,
    extra_offset=False,
    title="long",
    proj="stereo",
    show_points=True,
    show_surfaces=True,
    show=True,
    save_file=None,
    **kwargs,
):
    """
    Use projection to illustrate an EQ partition of S^3.

    Parameters
    ----------
    N : int
        Number of regions.
    extra_offset : bool, optional
        Use extra offsets. Default False.
    title : {'long', 'short', 'none'}, optional
        Title format. Default 'long'.
    proj : {'stereo', 'eqarea'}, optional
        Projection type. Default 'stereo'.
    show_points : bool, optional
        Show center points. Default True.
    show_surfaces : bool, optional
        Show region surfaces. Default True.
    **kwargs
        Passed to Mayavi plotting functions.

    Examples
    --------
    >>> from eqsp.visualizations import project_s3_partition
    >>> from mayavi import mlab
    >>> mlab.options.offscreen = True
    >>> try:
    ...     project_s3_partition(
    ...         4, proj='stereo', show_points=True, show_surfaces=False
    ...     )
    ...     print("Success")
    ... except ImportError:
    ...     print("Mayavi not installed")
    Success
    """
    if proj == "stereo":
        projector = x2stereo
    elif proj == "eqarea":
        projector = x2eqarea
    else:
        raise ValueError("proj must be 'stereo' or 'eqarea'")

    show_title = title != "none"

    # Set default figure size if none exists
    if mlab.get_engine().current_scene is None:
        mlab.figure(bgcolor=(1, 1, 1), size=(800, 800))

    dim = 3

    if show_surfaces:
        # Note: Extra offsets for Dim 3 not fully ported
        # (needs rotation matrices return from eq_regions)
        R = eq_regions(dim, N, extra_offset)

        for i in range(1, N):
            region = R[:, :, i]
            dim_reg = 3
            t = region[:, 0]
            b = region[:, 1]
            if abs(b[0]) < 1e-10:
                b[0] = 2 * np.pi  # pragma: no cover
            pseudo = abs(t[0]) < 1e-10 and abs(b[0] - 2 * np.pi) < 1e-10

            for k in range(dim_reg):
                if pseudo and k >= 2:
                    continue
                j = np.arange(dim_reg)
                j = np.roll(j, -k)

                h_grid = np.linspace(0, 1, 10)
                H1, H2 = np.meshgrid(h_grid, h_grid)

                s_face = np.zeros((dim_reg, 10, 10))
                idx_vary1, idx_vary2, idx_fixed = j[0], j[1], j[2]

                s_face[idx_vary1, :, :] = (
                    t[idx_vary1] + (b[idx_vary1] - t[idx_vary1]) * H1
                )
                s_face[idx_vary2, :, :] = (
                    t[idx_vary2] + (b[idx_vary2] - t[idx_vary2]) * H2
                )
                s_face[idx_fixed, :, :] = t[idx_fixed]

                s_flat = s_face.reshape(dim_reg, -1)
                x_flat = polar2cart(s_flat)
                p_flat = projector(x_flat)

                PX = p_flat[0, :].reshape(10, 10)
                PY = p_flat[1, :].reshape(10, 10)
                PZ = p_flat[2, :].reshape(10, 10)

                # Check for NaNs (e.g. projection to infinity)
                if np.any(np.isnan(PX)):
                    continue  # pragma: no cover

                # Mimic Matlab: color based on t[2] (jet), opacity = (t[2]/pi)/2
                cmap = plt.get_cmap("jet")
                c_val = t[2] / np.pi
                rgba = cmap(c_val)
                color = rgba[:3]
                opacity = (t[2] / np.pi) / 2.0

                mlab.mesh(PX, PY, PZ, opacity=opacity, color=color)

    if show_points:
        points = eq_point_set(dim, N, extra_offset)
        project_point_set(
            points, proj=proj, color=(1, 0, 0), scale_factor=0.1, show=False, **kwargs
        )

    if show_title:
        title_text = f"EQ(3,{N}) {PROJ_NAME.get(proj, proj)} projection"
        # width=0.6 gives a reasonable size for this longer string
        # x = 0.5 - 0.6/2 = 0.2, y = 0.9
        mlab.text(0.2, 0.9, title_text, width=0.6, color=(0, 0, 0))

    if save_file:
        mlab.savefig(save_file)

    if show:
        mlab.show()



"""
PyEQSP illustrations module.

Copyright 2026 Paul Leopardi
"""

import matplotlib.pyplot as plt
import numpy as np

from ._private._partitions import (
    cap_colats,
    ideal_region_list,
    num_collars,
    polar_colat,
    round_to_naturals,
)
from .partitions import eq_caps, eq_point_set, eq_regions
from .utilities import (
    ideal_collar_angle,
    polar2cart,
    x2eqarea,
    x2stereo,
)


def project_point_set(
    points, ax=None, proj="stereo", _scale_factor=0.05, color=None, show=None, **_kwargs
):
    """
    Use projection to illustrate a point set of S^2 or S^3.

    Parameters
    ----------
    points : ndarray
        Points in R^3 (S^2) or R^4 (S^3).
    ax : Axes, optional
        Matplotlib axes (2D or 3D). If None, a new figure is created.
    proj : {'stereo', 'eqarea'}, optional
        Projection type. Default is 'stereo'.
    _scale_factor : float, optional
        Scale factor for points (unused in matplotlib implementation, kept for
        compatibility/kwargs).
    color : color spec, optional
        Colour of points. Default is 'k' for 2D, 'r' for 3D.
    show : bool or None, optional
        If True, call `plt.show()`. If None, call `plt.show()` only if `ax` was None.
    **_kwargs
        Passed to `ax.scatter`.

    Returns
    -------
    ax : Axes
        The axes object.

    Examples
    --------
    >>> from eqsp.illustrations import project_point_set
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> plt.switch_backend('Agg')
    >>> points = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]).T
    >>> ax = project_point_set(points, proj='eqarea')
    >>> len(ax.collections)  # doctest: +SKIP
    1
    """
    points = np.asarray(points)
    dim = points.shape[0] - 1
    if dim not in [2, 3]:
        raise ValueError("Points must be in R^3 (S^2) or R^4 (S^3)")

    if proj == "stereo":
        projector = x2stereo
    elif proj == "eqarea":
        projector = x2eqarea
    else:
        raise ValueError("proj must be 'stereo' or 'eqarea'")

    if dim == 2:
        t = projector(points)
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.set_aspect("equal")
            ax.set_axis_off()
        if color is None:
            # Colour based on colatitude (mimic MATLAB)
            # Matlab uses r = pi - acos(z)
            z = points[2, :]
            r = np.pi - np.arccos(np.clip(z, -1.0, 1.0))
            cmap = plt.get_cmap("jet")
            c = cmap(r / np.pi)
        else:
            c = color
        ax.scatter(t[0, :], t[1, :], s=20, c=c, **_kwargs)

    elif dim == 3:
        raise NotImplementedError(
            "3D plotting for S^3 has been moved to eqsp.visualizations."
        )

    if show is True or (show is None and ax is None):
        if plt.get_backend().lower() != "agg":  # pragma: no cover
            plt.show()  # pragma: no cover
    return ax


def project_s2_partition(
    N,
    *,
    extra_offset=False,
    fontsize=16,
    title="long",
    proj="stereo",
    show_points=False,
    ax=None,
    show=None,
    **_kwargs,
):
    """
    Use projection to illustrate an EQ partition of S^2.

    Parameters
    ----------
    N : int
        Number of regions.
    extra_offset : bool, optional
        Use extra offsets. Default False.
    fontsize : int, optional
        Font size for title. Default 16.
    title : {'long', 'short', 'none'}, optional
        Title mode. Default 'long'.
    proj : {'stereo', 'eqarea'}, optional
        Projection type. Default 'stereo'.
    show_points : bool, optional
        Show center points of regions. Default False.
    ax : Axes, optional
        Matplotlib axes to plot on. If None, a new figure is created.
    show : bool or None, optional
        If True, call `plt.show()`. If None, call `plt.show()` only if `ax` was None.
    **_kwargs
        Passed to underlying plot functions.

    Returns
    -------
    ax : Axes
        The axes object.

    Examples
    --------
    >>> from eqsp.illustrations import project_s2_partition
    >>> import matplotlib.pyplot as plt
    >>> plt.switch_backend('Agg')
    >>> ax = project_s2_partition(4, proj='eqarea', title='none', show_points=True)
    >>> ax.get_title()  # doctest: +SKIP
    ''
    """
    show_title = title != "none"

    if proj == "stereo":
        projector = x2stereo
    elif proj == "eqarea":
        projector = x2eqarea
    else:
        raise ValueError("proj must be 'stereo' or 'eqarea'")

    dim = 2
    R = eq_regions(dim, N, extra_offset)

    if ax is None:
        fig = plt.figure(figsize=(8, 8))
        ax = fig.add_subplot(111)
        ax.set_aspect("equal")
        ax.set_axis_off()

    # Ensure aspect is equal even if ax provided
    ax.set_aspect("equal")
    ax.set_axis_off()

    if proj == "eqarea":
        circ = plt.Circle((0, 0), np.sqrt(2), color="k", fill=False, linewidth=1)
        ax.add_patch(circ)

    for i in range(1, N):  # Draw all regions 1..N-1?
        region = R[:, :, i]
        t = region[:, 0]
        b = region[:, 1]

        tol = 1e-10
        if abs(b[0]) < tol:
            b[0] = 2 * np.pi
        pseudo = abs(t[0]) < tol and abs(b[0] - 2 * np.pi) < tol

        fidelity = 33
        h = np.linspace(0, 1, fidelity)

        # Color: Black boundaries
        color = "k"

        for k in range(dim):
            if pseudo and k >= 1:
                continue
            j = np.arange(dim)
            j = np.roll(j, -k)

            s_curve = np.zeros((dim, fidelity))
            idx_vary = j[0]
            idx_fixed = j[1:]

            s_curve[idx_vary, :] = t[idx_vary] + (b[idx_vary] - t[idx_vary]) * h
            for i_f in idx_fixed:
                s_curve[i_f, :] = t[i_f]

            x_curve = polar2cart(s_curve)
            p_curve = projector(x_curve)

            mask = np.isfinite(p_curve[0, :])
            ax.plot(p_curve[0, mask], p_curve[1, mask], color=color, linewidth=0.5)

    if show_points:
        points = eq_point_set(dim, N, extra_offset)
        project_point_set(points, ax=ax, proj=proj, color=None, **_kwargs)

    if show_title:
        title_text = f"EQ(2,{N}) {proj} projection"
        ax.set_title(title_text, fontsize=fontsize, color="k")

    if show is True or (show is None and ax is None):
        if plt.get_backend().lower() != "agg":  # pragma: no cover
            plt.show()  # pragma: no cover

    return ax


def illustrate_eq_algorithm(
    dim,
    N,
    *,
    extra_offset=False,
    fontsize=8,
    show_title=True,
    _long_title=False,
    show_points=True,
    proj="eqarea",
    show=True,
    **_kwargs,
):
    """
    Illustrate the EQ partition algorithm.

    Parameters
    ----------
    dim : int
        Dimension of the sphere S^dim (unit sphere in R^{dim+1}).
    N : int
        Number of equal-area regions to partition the sphere into.
    extra_offset : bool, optional
        Use extra offsets. Default False.
    fontsize : int, optional
        Font size. Default 16.
    show_title : bool, optional
        Show title. Default True.
    _long_title : bool, optional
        Use long title variant. Default False.
    proj : str, optional
        Projection type ('stereo' or 'eqarea'). Default 'eqarea'.
    show : bool, optional
        If True (default), call `plt.show()`.
    **_kwargs
        Passed to underlying plot functions.

    Returns
    -------
    None
    """
    # Consolidate options for easier passing
    opts = {
        "extra_offset": extra_offset,
        "fontsize": fontsize,
        "show_title": show_title,
        "long_title": _long_title,  # Pass the original value, not the prefixed name
    }

    plt.subplot(2, 2, 1)
    plt.axis("off")
    illustrate_steps_1_2(dim, N, **opts)

    plt.subplot(2, 2, 2)
    plt.axis("off")
    illustrate_steps_3_5(dim, N, **opts)

    plt.subplot(2, 2, 3)
    plt.axis("off")
    illustrate_steps_6_7(dim, N, **opts)

    plt.subplot(2, 2, 4)
    plt.axis("off")
    plt.cla()

    # Increase font size for lower-dim visualisations
    proj_opts = opts.copy()
    proj_opts["fontsize"] = 16
    # Map proj options to illustration functions
    proj_opts["proj"] = proj

    proj_opts["show_points"] = show_points

    # Map title mode
    if show_title:
        proj_opts["title"] = "long" if _long_title else "short"
    else:
        proj_opts["title"] = "none"

    # Remove options not accepted by project_s2_partition/ax.scatter
    proj_opts.pop("show_title", None)
    proj_opts.pop("long_title", None)  # Remove the original name

    if dim == 2:
        ax = plt.subplot(2, 2, 4)
        plt.axis("off")
        project_s2_partition(N, ax=ax, **proj_opts)
    elif dim == 3:
        _, m = eq_caps(dim, N)
        # m is n_regions (1D array)
        max_collar = min(4, m.size - 2)
        plt.subplot(2, 2, 4)
        plt.axis("off")
        for k in range(1, max_collar + 1):
            subn = 9 + 2 * k - ((k - 1) % 2)
            ax_sub = plt.subplot(4, 4, subn)
            plt.axis("off")
            opts_k = proj_opts.copy()
            opts_k["title"] = "none"
            project_s2_partition(int(m[k]), ax=ax_sub, **opts_k)

    if show and plt.get_backend().lower() != "agg":
        plt.show()  # pragma: no cover


def illustrate_steps_1_2(
    dim, N, *, fontsize=8, show_title=True, _long_title=False, **_kwargs
):
    """
    Illustrate steps 1 and 2 of the EQ partition.
    """
    h = np.linspace(0.0, 1.0, 91)
    Phi = h * 2.0 * np.pi
    plt.plot(np.sin(Phi), np.cos(Phi), "k", linewidth=1)
    plt.axis("equal")
    plt.axis("off")

    c_polar = polar_colat(dim, N)

    k = np.linspace(-1.0, 1.0, 41)
    j = np.ones_like(k)

    # Plot the bounding parallels of the polar caps
    plt.plot(np.sin(c_polar) * k, np.cos(c_polar) * j, "r", linewidth=2)
    plt.plot(np.sin(c_polar) * k, -np.cos(c_polar) * j, "r", linewidth=2)

    # Plot the North-South axis
    plt.plot(np.zeros_like(j), k, "b", linewidth=1)
    # Plot the polar angle
    plt.plot(np.sin(c_polar) * h, np.cos(c_polar) * h, "b", linewidth=2)

    plt.text(0.05, 2.0 / 3.0, r"$\theta_c$", fontsize=fontsize)

    # Plot the ideal collar angle
    Delta_I = ideal_collar_angle(dim, N)
    theta = c_polar + Delta_I
    plt.plot(np.sin(theta) * h, np.cos(theta) * h, "b", linewidth=2)

    mid = c_polar + Delta_I / 2.0
    plt.text(
        np.sin(mid) * 2.0 / 3.0,
        np.cos(mid) * 2.0 / 3.0,
        r"$\Delta_I$",
        fontsize=fontsize,
    )

    # Plot an arc to indicate angles
    theta = h * (c_polar + Delta_I)
    plt.plot(np.sin(theta) / 5.0, np.cos(theta) / 5.0, "b", linewidth=1)

    plt.text(
        -0.9,
        -0.1,
        r"$V(\theta_c) = V_R$" + "\n" + r"$= \sigma(S^{%d})/%d$" % (dim, N),
        fontsize=fontsize,
    )

    caption_angle = min(mid + 2.0 * Delta_I, np.pi - c_polar)
    plt.text(
        np.sin(caption_angle) / 3.0,
        np.cos(caption_angle) / 3.0,
        r"$\Delta_I = V_R^{1/%d}$" % dim,
        fontsize=fontsize,
    )

    if show_title:
        title_str = f"EQ({dim},{N}) Steps 1 to 2"
        plt.title(title_str, fontsize=fontsize, color="k")


def illustrate_steps_3_5(
    dim, N, *, fontsize=8, show_title=True, _long_title=False, **_kwargs
):
    """
    Illustrate steps 3 to 5 of the EQ partition.
    """
    h = np.linspace(0.0, 1.0, 91)
    Phi = h * 2.0 * np.pi
    plt.plot(np.sin(Phi), np.cos(Phi), "k", linewidth=1)
    plt.axis("equal")
    plt.axis("off")

    c_polar = polar_colat(dim, N)
    n_collars = num_collars(N, c_polar, ideal_collar_angle(dim, N))
    r_regions = ideal_region_list(dim, N, c_polar, n_collars)
    s_cap = cap_colats(dim, N, c_polar, r_regions)

    k = np.linspace(-1.0, 1.0, 41)
    j = np.ones_like(k)
    plt.plot(np.sin(c_polar) * k, np.cos(c_polar) * j, "r", linewidth=2)
    plt.text(
        np.sin(c_polar) * 1.1,
        np.cos(c_polar) * 1.1,
        r"$\theta_{F,1}$",
        fontsize=fontsize,
    )
    plt.plot(np.sin(c_polar) * h, np.cos(c_polar) * h, "b", linewidth=2)
    plt.plot(np.sin(c_polar) * k, -np.cos(c_polar) * j, "r", linewidth=2)

    plt.plot(np.zeros_like(j), k, "b", linewidth=1)

    for collar_n in range(1, n_collars + 1):
        zone_n = collar_n
        theta = s_cap[zone_n]
        plt.plot(np.sin(theta) * h, np.cos(theta) * h, "b", linewidth=2)
        theta_str = r"$\theta_{F,%d}$" % (collar_n + 1)
        plt.text(np.sin(theta) * 1.1, np.cos(theta) * 1.1, theta_str, fontsize=fontsize)

        theta_p = s_cap[collar_n - 1]

        if collar_n > 1:
            plt.plot(np.sin(theta_p) * k, np.cos(theta_p) * j, "r", linewidth=2)

        arc = theta_p + (theta - theta_p) * h
        plt.plot(np.sin(arc) / 5.0, np.cos(arc) / 5.0, "b", linewidth=1)
        mid = (theta_p + theta) / 2.0
        plt.text(np.sin(mid) / 2.0, np.cos(mid) / 2.0, r"$\Delta_F$", fontsize=fontsize)
        y_str = r"$y_{%d} = %3.1f...$" % (collar_n, r_regions[zone_n])
        plt.text(
            -np.sin(mid) + 1.0 / 20.0,
            np.cos(mid) + (mid - np.pi) / 30.0,
            y_str,
            fontsize=fontsize,
        )
    if show_title:
        title_str = f"EQ({dim},{N}) Steps 3 to 5"
        plt.title(title_str, fontsize=fontsize, color="k")


def illustrate_steps_6_7(
    dim, N, *, fontsize=8, show_title=True, _long_title=False, **_kwargs
):
    """
    Illustrate steps 6 and 7 of the EQ partition.
    """
    h = np.linspace(0.0, 1.0, 91)
    Phi = h * 2.0 * np.pi
    plt.plot(np.sin(Phi), np.cos(Phi), "k", linewidth=1)
    plt.axis("equal")
    plt.axis("off")

    c_polar = polar_colat(dim, N)
    n_collars = num_collars(N, c_polar, ideal_collar_angle(dim, N))
    r_regions = ideal_region_list(dim, N, c_polar, n_collars)
    n_regions = round_to_naturals(N, r_regions)
    s_cap = cap_colats(dim, N, c_polar, n_regions)

    k = np.linspace(-1.0, 1.0, 41)
    j = np.ones_like(k)
    plt.plot(np.sin(c_polar) * k, np.cos(c_polar) * j, "r", linewidth=2)
    plt.text(
        np.sin(c_polar) * 1.1, np.cos(c_polar) * 1.1, r"$\theta_1$", fontsize=fontsize
    )
    plt.plot(np.sin(c_polar) * h, np.cos(c_polar) * h, "b", linewidth=2)
    plt.plot(np.sin(c_polar) * k, -np.cos(c_polar) * j, "r", linewidth=2)

    plt.plot(np.zeros_like(j), k, "b", linewidth=1)

    for collar_n in range(1, n_collars + 1):
        zone_n = collar_n
        theta = s_cap[zone_n]
        plt.plot(np.sin(theta) * h, np.cos(theta) * h, "b", linewidth=2)
        theta_str = r"$\theta_{%d}$" % (collar_n + 1)
        plt.text(np.sin(theta) * 1.1, np.cos(theta) * 1.1, theta_str, fontsize=fontsize)

        theta_p = s_cap[collar_n - 1]

        if collar_n > 1:
            plt.plot(np.sin(theta_p) * k, np.cos(theta_p) * j, "r", linewidth=2)

        arc = theta_p + (theta - theta_p) * h
        plt.plot(np.sin(arc) / 5.0, np.cos(arc) / 5.0, "b", linewidth=1)
        mid = (theta_p + theta) / 2.0
        Delta_str = r"$\Delta_{%i}$" % collar_n
        plt.text(np.sin(mid) / 2.0, np.cos(mid) / 2.0, Delta_str, fontsize=fontsize)
        m_str = r"$m_{%d} =%3.0f$" % (collar_n, n_regions[zone_n])
        plt.text(
            -np.sin(mid) + 1.0 / 20.0,
            np.cos(mid) + (mid - np.pi) / 30.0,
            m_str,
            fontsize=fontsize,
        )
    if show_title:
        title_str = f"EQ({dim},{N}) Steps 6 to 7"
        plt.title(title_str, fontsize=fontsize, color="k")



"""
PyEQSP region properties module.

Copyright 2026 Paul Leopardi
"""

import numpy as np

from ._private._region_props import (
    max_diam_bound_of_regions,
    max_vertex_diam_of_regions,
)
from .partitions import eq_regions
from .utilities import (
    area_of_collar,
    area_of_ideal_region,
    area_of_sphere,
)

__all__ = [
    "area_of_region",
    "eq_area_error",
    "eq_diam_bound",
    "eq_diam_coeff",
    "eq_regions_property",
    "eq_vertex_diam",
    "eq_vertex_diam_coeff",
]


def eq_area_error(dim, N, show_progress=False, even_collars=False):
    """
    Total area error and max area error per region of an EQ partition.

    Parameters
    ----------
    dim : int
        Dimension of the sphere.
    N : int or array-like of int
        Number of regions to partition.
    show_progress : bool, optional
        Show progress messages. Default False.
    even_collars : bool, optional
        Use even number of collars for symmetric partitions. Default False.

    Returns
    -------
    total_error : ndarray
        Absolute difference between total area of all regions and the area
        of $S^{dim}$. Total error should be near 0.
    max_error : ndarray
        Maximum absolute difference between the area of any region and the
        ideal area.

    Raises
    ------
    ValueError
        If arguments are not provided.

    See Also
    -------
    eq_regions, area_of_sphere, area_of_ideal_region

    Notes
    -----
    The results will be arrays of the same size as N.
    Note that both total_error and max_error are returned as native floats
    if N is a scalar, otherwise as NumPy arrays.

    Implementation
    --------------
    To accurately measure the accumulated rounding error of the partitioning
    algorithm itself, this function computes the area for every single region
    independently based on the exact geometric boundaries (`s_top` and `s_bot`)
    produced by the algorithm. It does not assume regions within a given collar
    are strictly identical, nor does it substitute theoretical ideal areas.
    The calculations are vectorized using NumPy arrays for performance while
    maintaining strict geometric fidelity to the algorithm's actual output.

    Examples
    --------
    >>> import numpy as np
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> total_error, max_error = eq_area_error(2, 10)
    >>> np.allclose(total_error, 0)
    True
    >>> total_error, max_error = eq_area_error(2, 6, show_progress=True)
    >>> float(total_error)
    0.0
    >>> np.allclose(max_error, 0, atol=1e-12)
    True
    >>> np.allclose(total_error * 1e12, [0, 0, 0, 0, 0, 0], atol=1)
    True
    >>> np.allclose(max_error * 1e12, [0, 0, 0, 0, 0, 0], atol=1)
    True
    >>> total_error, max_error = eq_area_error(2, 6, show_progress=True)
    >>> np.allclose(total_error, 0)
    True
    """
    if dim is None or N is None:
        raise ValueError("Both dim and N must be provided.")

    shape = np.shape(N)
    n_partitions = int(np.prod(shape))
    N_flat = np.reshape(N, (1, n_partitions))

    total_error = np.zeros_like(N_flat, dtype=float)
    max_error = np.zeros_like(N_flat, dtype=float)
    sphere_area = area_of_sphere(dim)

    for i in range(n_partitions):
        n = int(N_flat[0, i])
        if show_progress and n_partitions > 1:
            print(f"    N={n:6} ({i + 1}/{n_partitions})", end="\r", flush=True)
        regions = eq_regions(dim, n, even_collars=even_collars)
        ideal_area = area_of_ideal_region(dim, n)

        areas = area_of_region(regions)
        total_area = np.sum(areas)

        # max_error logic
        region_errors = np.abs(areas - ideal_area)
        if np.size(region_errors) > 0:
            max_error[0, i] = np.max(region_errors)

        total_error[0, i] = abs(sphere_area - total_area)

    if show_progress and n_partitions > 1:
        print()  # Clear the line
    total_error = np.reshape(total_error, shape)
    max_error = np.reshape(max_error, shape)
    return total_error, max_error


def area_of_region(region):
    """
    Area of given region(s).

    Parameters
    ----------
    region : ndarray
        Region(s), typically of shape (dim, 2) or (dim, 2, N).

    Returns
    -------
    area : float or ndarray
        Area of the region(s).

    See Also
    --------
    area_of_sphere, area_of_collar

    Examples
    --------
    >>> import numpy as np
    >>> region = np.array([[0, 2*np.pi], [0, np.pi]])
    >>> float(area_of_region(region))
    12.566370614359172
    """
    dim = region.shape[0]
    s_top = region[dim - 1, 0, ...]
    s_bot = region[dim - 1, 1, ...]
    if dim > 1:
        area = (
            area_of_collar(dim, s_top, s_bot)
            * area_of_region(region[: dim - 1, ...])
            / area_of_sphere(dim - 1)
        )
    else:
        if np.ndim(s_bot) == 0:
            if s_bot == 0:
                s_bot = 2 * np.pi
            if s_top == s_bot:
                s_bot = s_top + 2 * np.pi
            area = s_bot - s_top
        else:
            s_bot = np.copy(s_bot)
            s_bot[s_bot == 0] = 2 * np.pi
            mask = s_top == s_bot
            s_bot[mask] = s_top[mask] + 2 * np.pi
            area = s_bot - s_top
    return area


def eq_diam_bound(dim, N, show_progress=False, even_collars=False):
    """
    Upper bound on diameter of regions of an EQ partition.

    Parameters
    ----------
    dim : int
        Dimension of the sphere.
    N : int or array-like
        Number of regions.
    show_progress : bool, optional
        Show progress messages. Default False.
    even_collars : bool, optional
        Use even number of collars for symmetric partitions. Default False.

    Returns
    -------
    diam_bound : ndarray
        Maximum of per-region diameter bound.

    See Also
    --------
    eq_vertex_diam, eq_diam_coeff, eq_regions_property

    Examples
    --------
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> print(f"{float(eq_diam_bound(2, 10)):.4f}")
    1.6733
    >>> eq_diam_bound(3, np.arange(1, 7))
    array([2., 2., 2., 2., 2., 2.])
    >>> print(f"{float(eq_diam_bound(2, 6, show_progress=True)):.4f}")
    1.9437
    """
    return eq_regions_property(
        max_diam_bound_of_regions,
        dim,
        N,
        show_progress=show_progress,
        even_collars=even_collars,
    )


def eq_diam_coeff(dim, N, show_progress=False, even_collars=False):
    """
    Coefficient of diameter bound for EQ partitions.

    Parameters
    ----------
    dim : int
        Dimension of the sphere.
    N : int or array-like
        Number of regions.
    show_progress : bool, optional
        Show progress messages. Default False.
    even_collars : bool, optional
        Use even number of collars for symmetric partitions. Default False.

    Returns
    -------
    bound_coeff : ndarray
        Diameter bound coefficient.
    vertex_coeff : ndarray
        Vertex diameter coefficient.

    Raises
    ------
    ValueError
        If arguments are not provided.

    See Also
    --------
    eq_diam_bound, eq_vertex_diam, eq_regions, eq_vertex_diam_coeff

    Notes
    -----
    The returned coefficients relate the maximum per-region diameter to the
    number of regions. Specifically, the coefficients are calculated as:
    :math:`C = \\text{diam} \\times N^{1/d}`
    where ``diam`` is the maximum diameter bound or vertex diameter.

    Unlike the Matlab implementation, this function always returns a 2-tuple.
    If you only need one coefficient, you can ignore the other during unpacking
    (e.g., ``b, _ = eq_diam_coeff(...)``).

    Examples
    --------
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> bound_coeff, vertex_coeff = eq_diam_coeff(2, 10)
    >>> print(f"{float(bound_coeff):.4f}, {float(vertex_coeff):.4f}")
    5.2915, 4.4721
    >>> b, v = eq_diam_coeff(2, 6, show_progress=True)
    >>> print(f"{float(b):.4f}")
    4.7610
    """
    if dim is None or N is None:
        raise ValueError("Both dim and N must be provided.")

    shape = np.shape(N)
    n_partitions = int(np.prod(shape))
    N_flat = np.reshape(N, (1, n_partitions))
    bound_coeff = np.zeros_like(N_flat, dtype=float)
    vertex_coeff = np.zeros_like(N_flat, dtype=float)
    for i in range(n_partitions):
        n = int(N_flat[0, i])
        if show_progress and n_partitions > 1:
            print(f"    N={n:6} ({i + 1}/{n_partitions})", end="\r", flush=True)
        regions = eq_regions(dim, n, even_collars=even_collars)
        scale = np.power(n, 1 / dim)
        bound_coeff[0, i] = max_diam_bound_of_regions(regions) * scale
        vertex_coeff[0, i] = max_vertex_diam_of_regions(regions) * scale
    if show_progress and n_partitions > 1:
        print()  # Clear the line
    bound_coeff = np.reshape(bound_coeff, shape)
    vertex_coeff = np.reshape(vertex_coeff, shape)
    return bound_coeff, vertex_coeff


def eq_regions_property(fhandle, dim, N, show_progress=False, even_collars=False):
    """
    Property of regions of an EQ partition.

    Parameters
    ----------
    fhandle : callable
        Function handle to apply to regions.
    dim : int
        Dimension of sphere.
    N : int or array-like of int
        Number of regions.
    show_progress : bool, optional
        Show progress messages. Default False.
    even_collars : bool, optional
        Use even number of collars for symmetric partitions. Default False.

    Returns
    -------
    property : ndarray
        Calculated property for each partition.

    See Also
    --------
    eq_regions, eq_diam_bound, eq_vertex_diam

    Notes
    -----
    The function specified by fhandle must accept an array of shape
    (dim, 2, N_regions) and return a single value.

    Examples
    --------
    >>> def dummy_property(regions): return regions.shape[2]
    >>> eq_regions_property(dummy_property, 2, [3, 4]).astype(int)
    array([3, 4])
    >>> float(eq_regions_property(dummy_property, 2, 6, show_progress=True))
    6.0
    """
    shape = np.shape(N)
    n_partitions = int(np.prod(shape))
    N_flat = np.reshape(N, (1, n_partitions))
    property_ = np.zeros_like(N_flat, dtype=float)
    for i in range(n_partitions):
        n = int(N_flat[0, i])
        if show_progress and n_partitions > 1:
            print(f"    N={n:6} ({i + 1}/{n_partitions})", end="\r", flush=True)
        regions = eq_regions(dim, n, even_collars=even_collars)
        property_[0, i] = fhandle(regions)
    if show_progress and n_partitions > 1:
        print()  # Clear the line
    property_ = np.reshape(property_, shape)
    return property_


def eq_vertex_diam_coeff(dim, N, show_progress=False, even_collars=False):
    """
    Coefficient of maximum vertex diameter of regions of an EQ partition.

    Parameters
    ----------
    dim : int
        Dimension of the sphere.
    N : int or array-like
        Number of regions.
    show_progress : bool, optional
        Show progress messages. Default False.
    even_collars : bool, optional
        Use even number of collars for symmetric partitions. Default False.

    Returns
    -------
    coeff : ndarray
        Vertex diameter coefficient.

    See Also
    --------
    eq_vertex_diam, eq_diam_coeff

    Examples
    --------
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> print(f"{float(eq_vertex_diam_coeff(2, 10)):.4f}")
    4.4721
    >>> eq_vertex_diam_coeff(3, np.arange(1, 7))
    array([2.    , 2.5198, 2.8845, 3.1748, 3.42  , 3.6342])
    >>> print(f"{float(eq_vertex_diam_coeff(2, 6, show_progress=True)):.4f}")
    4.1633
    """
    diam = eq_vertex_diam(dim, N, show_progress, even_collars=even_collars)
    return diam * np.power(N, 1 / dim)


def eq_vertex_diam(dim, N, show_progress=False, even_collars=False):
    """
    Maximum vertex diameter of regions of an EQ partition.

    Parameters
    ----------
    dim : int
        Dimension of the sphere.
    N : int or array-like
        Number of regions.
    show_progress : bool, optional
        Show progress messages. Default False.
    even_collars : bool, optional
        Use even number of collars for symmetric partitions. Default False.

    Returns
    -------
    vertex_diam : ndarray
        Maximum vertex diameter over all regions.

    See Also
    --------
    eq_diam_bound, eq_vertex_diam_coeff, eq_diam_coeff, eq_regions_property

    Examples
    --------
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> print(f"{float(eq_vertex_diam(2, 10)):.4f}")
    1.4142
    >>> eq_vertex_diam(3, np.arange(1, 7))
    array([2., 2., 2., 2., 2., 2.])
    >>> print(f"{float(eq_vertex_diam(2, 6, show_progress=True)):.4f}")
    1.6997
    """
    return eq_regions_property(
        max_vertex_diam_of_regions,
        dim,
        N,
        show_progress=show_progress,
        even_collars=even_collars,
    )


if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod()



"""
PyEQSP histograms module.

Copyright 2026 Paul Leopardi
"""

import numpy as np

from ._private._histograms import lookup_s2_region
from .partitions import eq_caps, eq_regions


def eq_count_points_by_s2_region(s_point, N):
    """
    Count points in each of N regions of S^2.

    Parameters
    ----------
    s_point : ndarray
        Sequence of points on S^2, as a 2 x n_points array in spherical
        polar coordinates, with longitude 0 <= s[0, p_idx] <= 2*pi,
        colatitude 0 <= s[1, p_idx] <= pi.
    N : int
        Required number of regions, a positive integer.

    Returns
    -------
    count_v : ndarray
        Array of length N containing the number of points of s_point
        contained in each region.

    See Also
    --------
    eq_find_s2_region

    Examples
    --------
    >>> import numpy as np
    >>> from eqsp.partitions import eq_point_set_polar
    >>> points_s = eq_point_set_polar(2, 8)
    >>> eq_count_points_by_s2_region(points_s, 8)
    array([1, 1, 1, 1, 1, 1, 1, 1])
    >>> eq_count_points_by_s2_region(points_s, 5)
    array([1, 2, 2, 2, 1])
    >>> points_s = eq_point_set_polar(2, 128)
    >>> eq_count_points_by_s2_region(points_s, 8)
    array([19, 15, 14, 17, 15, 14, 15, 19])
    >>> eq_count_points_by_s2_region(points_s, 5)
    array([19, 29, 32, 29, 19])
    """
    r_idx = eq_find_s2_region(s_point, N)
    count_v = np.histogram(r_idx, bins=np.arange(1, N + 2))[0]
    return count_v


def eq_find_s2_region(s_point, N):
    """
    Partition S^2 into N regions and find the index for each point.

    Parameters
    ----------
    s_point : ndarray
        Sequence of points on S^2, as a 2 x n_points array in spherical
        polar coordinates, with longitude 0 <= s[0, p_idx] <= 2*pi,
        colatitude 0 <= s[1, p_idx] <= pi.
    N : int
        Required number of regions, a positive integer.

    Returns
    -------
    r_idx : ndarray
        Array of length s_point.shape[1] containing the index of the region
        corresponding to each point.

    See Also
    --------
    eq_count_points_by_s2_region
    lookup_s2_region

    Examples
    --------
    >>> from eqsp.partitions import eq_point_set_polar
    >>> points_s = eq_point_set_polar(2, 8)
    >>> eq_find_s2_region(points_s, 8)
    array([1, 2, 3, 4, 5, 6, 7, 8])
    >>> eq_find_s2_region(points_s, 5)
    array([1, 2, 2, 3, 3, 4, 4, 5])
    >>> # Testing an exact colatitude boundary (side='left' logic)
    >>> import numpy as np
    >>> from eqsp.partitions import eq_caps
    >>> cap_boundary = eq_caps(2, 5)[0][0]
    >>> points_on_boundary = np.array([[0.0, np.pi], [cap_boundary, cap_boundary]])
    >>> eq_find_s2_region(points_on_boundary, 5)
    array([1, 1])
    """
    s_regions = eq_regions(2, N)
    s_cap, n_regions = eq_caps(2, N)
    c_regions = np.cumsum(n_regions)
    r_idx = lookup_s2_region(s_point, s_regions, s_cap, c_regions)
    return r_idx


def in_s2_region(s_point, region):
    """
    Test if points on S^2 are within a given region.

    Parameters
    ----------
    s_point : ndarray
        Sequence of points on S^2, as a 2 x n_points array in spherical
        polar coordinates, with longitude 0 <= s[0, p_idx] <= 2*pi,
        colatitude 0 <= s[1, p_idx] <= pi.
    region : ndarray
        One region of S^2 as returned by eq_regions(2, N).

    Returns
    -------
    result : ndarray
        Boolean array of length s_point.shape[1] indicating whether each
        point is in the region.

    See Also
    --------
    eq_regions
    eq_find_s2_region

    Examples
    --------
    >>> from eqsp.partitions import eq_point_set_polar, eq_regions
    >>> points_s = eq_point_set_polar(2, 8)
    >>> s_regions = eq_regions(2, 5)
    >>> region = s_regions[:, :, 2]
    >>> in_s2_region(points_s, region)
    array([False, False, False,  True,  True, False, False, False])
    """
    if region.shape != (2, 2):
        raise ValueError(f"Region must have shape (2, 2), got {region.shape}")
    n_points = s_point.shape[1]
    result = np.zeros(n_points, dtype=bool)
    for p_idx in range(n_points):
        longitude = s_point[0, p_idx]
        min_long = region[0, 0]
        max_long = region[0, 1]
        in_long = min_long < longitude <= max_long
        if not in_long:
            longitude = longitude + 2 * np.pi
            in_long = min_long < longitude <= max_long
        colatitude = s_point[1, p_idx]
        min_colat = region[1, 0]
        max_colat = region[1, 1]
        in_colat = True
        if (
            (min_colat == 0.0 and colatitude < min_colat)
            or (min_colat > 0.0 and colatitude <= min_colat)
            or (max_colat < colatitude)
        ):
            in_colat = False
        result[p_idx] = in_long and in_colat
    return result



"""
PyEQSP: Python Equal Area Sphere Partitioning Library.

This module provides functions for equal area sphere partitioning, including:
- Creating partitions (`eq_regions`, `eq_point_set`, `eq_caps`)
- Calculating properties (`point_set_props`, `region_props`)
- Utilities (`utilities`)
- Illustrations (`illustrations`)
"""

from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as _pkg_version

from eqsp.partitions import eq_caps, eq_point_set, eq_point_set_polar, eq_regions
from eqsp.point_set_props import (
    eq_dist_coeff,
    eq_energy_coeff,
    eq_energy_dist,
    eq_min_dist,
    point_set_dist_coeff,
    point_set_energy_coeff,
    point_set_energy_dist,
    point_set_min_dist,
)
from eqsp.region_props import eq_diam_bound
from eqsp.utilities import (
    area_of_cap,
    area_of_collar,
    area_of_sphere,
    cart2polar2,
    polar2cart,
    volume_of_ball,
)

__all__ = [
    "eq_regions",
    "eq_point_set",
    "eq_point_set_polar",
    "eq_caps",
    "volume_of_ball",
    "area_of_sphere",
    "area_of_cap",
    "area_of_collar",
    "polar2cart",
    "cart2polar2",
    "eq_min_dist",
    "eq_dist_coeff",
    "eq_energy_dist",
    "eq_energy_coeff",
    "eq_diam_bound",
    "point_set_min_dist",
    "point_set_energy_dist",
    "point_set_dist_coeff",
    "point_set_energy_coeff",
]

try:
    __version__ = _pkg_version("pyeqsp")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



def lookup_s2_region(s_point, s_regions, s_cap, c_regions):
    """
    For S^2, given sequences of points, regions, and cap colatitudes, find the
    index of the region containing each point.

    Parameters
    ----------
    s_point : ndarray
        Sequence of points on S^2, as a 2 x n_points array in spherical polar
        coordinates, with longitude 0 <= s[0, p_idx] <= 2 * pi, colatitude
        0 <= s[1, p_idx] <= pi.
    s_regions : ndarray
        Sequence of regions of S^2 as per eq_regions(2, N) where N ==
        s_regions.shape[2].
    s_cap : ndarray
        Sequence of cap colatitudes as per eq_caps(2, N) for the same N.
    c_regions : ndarray
        Sequence of the cumulative number of regions of s_regions within each
        cap of s_cap.

    Returns
    -------
    r_idx : ndarray
        Array of length s_point.shape[1] containing the index of the region of
        s_regions corresponding to each point.

    Raises
    ------
    ValueError
        If input dimensions are inconsistent.

    See Also
    --------
    eq_regions, eq_caps, numpy.cumsum, lookup_table

    Notes
    -----
    For references, see AUTHORS. For revision history, see CHANGELOG.

    Examples
    --------
    >>> import numpy as np
    >>> from eqsp._private._histograms import lookup_s2_region
    >>> import eqsp
    >>> points_s = eqsp.eq_point_set_polar(2, 8)
    >>> N = 8
    >>> s_regions = eqsp.eq_regions(2, N)
    >>> s_cap, n_regions = eqsp.eq_caps(2, N)
    >>> c_regions = np.cumsum(n_regions)
    >>> r_idx = lookup_s2_region(points_s, s_regions, s_cap, c_regions)
    >>> print(r_idx)
    [1 2 3 4 5 6 7 8]
    >>> N = 5
    >>> s_regions = eqsp.eq_regions(2, N)
    >>> s_cap, n_regions = eqsp.eq_caps(2, N)
    >>> c_regions = np.cumsum(n_regions)
    >>> r_idx = lookup_s2_region(points_s, s_regions, s_cap, c_regions)
    >>> print(r_idx)
    [1 2 2 3 3 4 4 5]
    """
    import numpy as np

    n_caps = len(s_cap)
    if n_caps != len(c_regions):
        raise ValueError(
            "LOOKUP_S2_REGION: Mismatch between length of s_cap and c_regions"
        )
    # The last element of c_regions should be the total number of regions (N)
    n_regions = s_regions.shape[2]
    if c_regions[n_caps - 1] != n_regions:
        raise ValueError(
            "LOOKUP_S2_REGION: Mismatch between c_regions[-1] and length of s_regions"
        )
    n_points = s_point.shape[1]
    r_idx = np.zeros(n_points, dtype=int)
    if n_points == 0:
        return r_idx

    c_idx_all = np.atleast_1d(lookup_table(s_cap, s_point[1, :]))

    r_idx[c_idx_all == 0] = 1
    r_idx[c_idx_all >= n_caps - 1] = n_regions

    active_mask = (c_idx_all > 0) & (c_idx_all < n_caps - 1)
    if np.any(active_mask):
        active_c_idxs = c_idx_all[active_mask]
        active_longs = s_point[0, active_mask]
        orig_indices = np.where(active_mask)[0]

        for c_idx in np.unique(active_c_idxs):
            collar_mask = active_c_idxs == c_idx
            pts_long = active_longs[collar_mask]
            pts_idx = orig_indices[collar_mask]

            min_r_idx = int(c_regions[c_idx - 1]) + 1
            max_r_idx = int(c_regions[c_idx])

            s_longs = s_regions[0, :, min_r_idx - 1 : max_r_idx].copy()
            # Normalize to [0, 2*pi]
            s_longs %= 2 * np.pi
            if s_longs[1, 0] < s_longs[0, 0]:
                # This region wraps around 2*pi, e.g., [6.1, 0.2]
                # In lookup_table context, we want to look at it as [6.1, 6.4]
                # but following regions will be [0.2, 0.7] -> [6.4, 6.9]
                # This is complex, let's just ensure the ends are increasing
                pass  # pragma: no cover

            n_longs = s_longs.shape[1]
            if n_longs > 1:
                # If plural, ensure ends are monotonically increasing by adding 2*pi
                # wherever they jump down.
                ends = s_longs[1, :].copy()
                for i in range(1, n_longs):
                    while ends[i] < ends[i - 1]:
                        ends[i] += 2 * np.pi
                table = ends
            else:
                table = s_longs[1, :]

            l_idx = np.atleast_1d(lookup_table(table, pts_long)) % n_longs

            wrap_mask = pts_long < s_longs[0, 0]
            l_idx[wrap_mask] = n_longs - 1

            r_idx[pts_idx] = min_r_idx + l_idx

    return r_idx


def lookup_table(table, y):
    """
    Lookup values in a sorted table. Usually used as a prelude to interpolation.

    If table is strictly increasing and idx = lookup_table(table, y), then
    table[idx[i]] <= y[i] < table[idx[i]+1] for all y[i] within the interval
    with minimum table[0] and maximum table[-1]. If y[i] < table[0] then
    idx[i] is 0. If y[i] >= table[-1] then idx[i] is len(table) - 1.

    If the table is strictly decreasing, then the tests are reversed.
    (NOT YET IMPLEMENTED). There are no guarantees for tables which are
    non-monotonic or are not strictly monotonic.

    Parameters
    ----------
    table : array_like
        Sequence of real values, assumed to be strictly increasing or decreasing.
    y : array_like or float
        Sequence of real values to be looked up in table.

    Returns
    -------
    idx : int or ndarray
        Indices in table corresponding to each y.

    Raises
    ------
    NotImplementedError
        If table is strictly decreasing.

    See Also
    --------
    numpy.searchsorted

    Notes
    -----
    For references, see AUTHORS. For revision history, see CHANGELOG.

    Examples
    --------
    >>> table = [-100.0, -70, 2.5, 75, 125.7]
    >>> y = [-1, 3, 1000, -197]
    >>> lookup_table(table, y)
    array([2, 3, 4, 0])
    """
    import numpy as np

    table = np.asarray(table)
    y = np.atleast_1d(y)

    # Strictly decreasing table is NOT YET IMPLEMENTED.
    if len(table) > 1 and table[0] > table[-1]:
        raise NotImplementedError("lookup_table: Decreasing table NOT YET IMPLEMENTED")

    # Nondecreasing table.
    maximum = np.max(np.concatenate([table, y])) + 1
    extended_table = np.append(table, maximum)
    idx = np.searchsorted(extended_table, y, side="left")
    idx[idx < 0] = 0

    # Cap at len(table) - 1 per docstring and typical interpolation use cases
    n_table = len(table)
    idx[idx >= n_table] = n_table - 1

    if idx.size == 1:
        return int(idx[0])
    return idx



import numpy as np

from ..utilities import (
    euclidean_dist,
    polar2cart,
    sph2euc_dist,
)


def expand_region_for_diam(region):
    """
    The set of 2^d vertices of a region.

    Expands a region from the 2 vertex definition to the set of 2^dim vertices
    of the pseudo-region of a region, so that the Euclidean diameter of a region
    is approximated by the diameter of this set.

    Parameters
    ----------
    region : ndarray
        The input region, shape (dim, 2).

    Returns
    -------
    expanded_region : ndarray
        The expanded region, shape (dim, 2**dim).

    See Also
    --------
    pseudo_region_for_diam, max_vertex_diam_of_regions, eq_vertex_diam

    Examples
    --------
    >>> regions = eq_regions(2, 10)
    >>> region = regions[:, :, 1]
    >>> expanded_region = expand_region_for_diam(region)
    >>> np.round(expanded_region, 4)
    array([[0.    , 1.5708, 0.    , 1.5708],
           [0.6435, 0.6435, 1.5708, 1.5708]])
    """
    dim = region.shape[0]
    if dim > 1:
        s_top = region[dim - 1, 0]
        s_bot = region[dim - 1, 1]
        region_1 = expand_region_for_diam(region[: dim - 1, :])
        expanded_region = np.concatenate(
            [append(region_1, s_top), append(region_1, s_bot)], axis=1
        )
    else:
        expanded_region = pseudo_region_for_diam(region)
    return expanded_region


def append(matrix, value):
    """
    Append a coordinate value to each column of a matrix.

    Parameters
    ----------
    matrix : ndarray
        Input matrix, shape (n, m).
    value : float
        Value to append as a new row.

    Returns
    -------
    result : ndarray
        Matrix with appended row, shape (n+1, m).

    Examples
    --------
    >>> m = np.array([[1, 2], [3, 4]])
    >>> append(m, 5)
    array([[1., 2.],
           [3., 4.],
           [5., 5.]])
    """
    return np.vstack([matrix, np.ones(matrix.shape[1]) * value])


def max_diam_bound_of_regions(regions):
    """
    The maximum diameter bound in an array of regions.

    Parameters
    ----------
    regions : ndarray
        Array of regions, shape (dim, 2, N).

    Returns
    -------
    diam_bound : float
        The maximum diameter bound.

    See Also
    --------
    max_vertex_diam_of_regions, eq_diam_bound

    Notes
    -----
    For licensing, see COPYING.
    For references, see AUTHORS.
    For revision history, see CHANGELOG.

    Examples
    --------
    >>> regions = eq_regions(2, 10)
    >>> float(np.round(max_diam_bound_of_regions(regions), 4))
    1.6733
    """
    dim = regions.shape[0]
    if dim == 1:
        diam_bound = diam_bound_region(regions[:, :, 0])  # pragma: no cover
    else:
        colatitude = -np.inf * np.ones(dim - 1)
        diam_bound = 0.0
        N = regions.shape[2]
        for region_n in range(N):
            top = regions[:, 0, region_n]
            if np.linalg.norm(top[1:dim] - colatitude) != 0:
                colatitude = top[1:dim]
                diam_bound = max(diam_bound, diam_bound_region(regions[:, :, region_n]))
    return diam_bound


def diam_bound_region(region):
    """
    Calculate the per-region bound on the Euclidean diameter of a region.

    Parameters
    ----------
    region : ndarray
        Input region, shape (dim, 2).

    Returns
    -------
    diam_bound : float
        Per-region diameter bound.

    See Also
    --------
    pseudo_region_for_diam

    Examples
    --------
    >>> regions = eq_regions(2, 10)
    >>> float(np.round(diam_bound_region(regions[:, :, 1]), 4))
    1.6733
    """
    tol = np.finfo(float).eps * 2**5
    pseudo_region = pseudo_region_for_diam(region)
    dim = pseudo_region.shape[0]
    top = pseudo_region[:, 0]
    bot = pseudo_region[:, 1]
    s = bot[dim - 1] - top[dim - 1]
    e = sph2euc_dist(s)
    if dim == 1:
        diam_bound = e
    else:
        max_sin = max(np.sin(top[dim - 1]), np.sin(bot[dim - 1]))
        if (top[dim - 1] <= np.pi / 2) and (bot[dim - 1] >= np.pi / 2):
            max_sin = 1
        if (abs(top[dim - 1]) < tol) or (abs(np.pi - bot[dim - 1]) < tol):
            diam_bound = 2 * max_sin
        else:
            region_1 = np.column_stack([top[: dim - 1], bot[: dim - 1]])
            diam_bound_1 = max_sin * diam_bound_region(region_1)
            diam_bound = min(2, np.sqrt(e**2 + diam_bound_1**2))
    return diam_bound


def max_vertex_diam_of_regions(regions):
    """
    The max vertex diameter in a cell array of regions.

    Parameters
    ----------
    regions : ndarray
        Array of regions, shape (dim, 2, N).

    Returns
    -------
    vertex_diam : float
        The maximum vertex diameter.

    See Also
    --------
    max_diam_bound_of_regions, eq_vertex_diam

    Examples
    --------
    >>> regions = eq_regions(2, 10)
    >>> float(np.round(max_vertex_diam_of_regions(regions), 4))
    1.4142
    """
    dim = regions.shape[0]
    if dim == 1:
        vertex_diam = vertex_diam_region(regions[:, :, 0])  # pragma: no cover
    else:
        colatitude = -np.inf * np.ones(dim - 1)
        vertex_diam = 0.0
        N = regions.shape[2]
        for region_n in range(N):
            top = regions[:, 0, region_n]
            if np.linalg.norm(top[1:dim] - colatitude) != 0:
                colatitude = top[1:dim]
                vertex_diam = max(
                    vertex_diam, vertex_diam_region(regions[:, :, region_n])
                )
    return vertex_diam


def vertex_diam_region(region):
    """
    Calculate the Euclidean diameter of the set of 2^dim vertices
    of the pseudo-region of a region.

    Parameters
    ----------
    region : ndarray
        Input region, shape (dim, 2).

    Returns
    -------
    diam : float
        The diameter.

    See Also
    --------
    expand_region_for_diam

    Examples
    --------
    >>> regions = eq_regions(2, 10)
    >>> float(np.round(vertex_diam_region(regions[:, :, 1]), 4))
    1.4142
    """
    expanded_region = expand_region_for_diam(region)
    dim = expanded_region.shape[0]
    full = expanded_region.shape[1]
    half = full // 2
    top = expanded_region[:, 0]
    bot = expanded_region[:, full - 1]
    diam = 0.0
    if np.sin(top[dim - 1]) > np.sin(bot[dim - 1]):
        for point_n_1 in range(0, half, 2):
            for point_n_2 in range(point_n_1 + 1, full, 2):
                x1 = polar2cart(expanded_region[:, point_n_1])
                x2 = polar2cart(expanded_region[:, point_n_2])
                diam = max(diam, euclidean_dist(x1, x2))
    else:
        for point_n_1 in range(full - 1, half, -2):
            for point_n_2 in range(point_n_1 - 1, -1, -2):
                x1 = polar2cart(expanded_region[:, point_n_1])
                x2 = polar2cart(expanded_region[:, point_n_2])
                diam = max(diam, euclidean_dist(x1, x2))
    return diam


def pseudo_region_for_diam(region):
    """
    Two points which maximize the vertex diameter of a region.

    Parameters
    ----------
    region : ndarray
        Input region, shape (dim, 2).

    Returns
    -------
    pseudo_region : ndarray
        Pseudo-region, shape (dim, 2).

    See Also
    --------
    expand_region_for_diam, max_vertex_diam_of_regions, eq_vertex_diam

    Examples
    --------
    >>> regions = eq_regions(2, 10)
    >>> region = regions[:, :, 1]
    >>> pseudo_region = pseudo_region_for_diam(region)
    >>> np.round(pseudo_region, 4)
    array([[0.    , 1.5708],
           [0.6435, 1.5708]])
    """
    tol = np.finfo(float).eps * 2**5
    phi_top = region[0, 0]
    phi_bot = region[0, 1]
    if phi_bot == 0:
        phi_bot = 2 * np.pi
    if (
        np.mod(phi_bot - phi_top, 2 * np.pi) < tol
        or np.mod(phi_bot - phi_top, 2 * np.pi) > np.pi
    ):
        phi_bot = phi_top + np.pi
    pseudo_region = region.copy()
    pseudo_region[0, 1] = phi_bot
    return pseudo_region



"""
PyEQSP point set properties module.

Copyright 2026 Paul Leopardi
"""

import math

import numpy as np
from scipy.spatial import KDTree
from scipy.spatial.distance import cdist
from scipy.special import psi

from .partitions import eq_caps, eq_point_set
from .utilities import (
    area_of_cap,
    area_of_sphere,
    euc2sph_dist,
)

__all__ = [
    "calc_dist_coeff",
    "calc_energy_coeff",
    "calc_packing_density",
    "eq_dist_coeff",
    "eq_energy_coeff",
    "eq_energy_dist",
    "eq_min_dist",
    "point_set_dist_coeff",
    "point_set_energy_coeff",
    "point_set_energy_dist",
    "point_set_min_dist",
    "sphere_int_energy",
]


def calc_dist_coeff(dim, N, min_euclidean_dist):
    """
    Coefficient of minimum distance.

    Parameters
    ----------
    dim : int
        Number of dimensions, must be positive integer.
    N : int or array-like
        Number of regions, must be positive integer(s).
    min_euclidean_dist : array-like
        Minimum Euclidean distance(s), same shape as N.

    Returns
    -------
    coeff : array-like
        Coefficient(s), same shape as N.

    Notes
    -----
    The expression for the lower bound on minimum distance of a minimum
    $r^{-s}$ energy point set on $S^{dim}$ was given by
    :ref:`[Rak94] <rak94>` for s == 0, dim = 2,
    :ref:`[Dah78] <dah78>` for s == dim-1,
    :ref:`[Kui07] <kui07>` for dim-1 <= s < dim,
    and :ref:`[Kui98] <kui98>` for s > dim.

    See Also
    --------
    eq_min_dist, eq_dist_coeff

    Examples
    --------
    >>> N = np.arange(2, 7)
    >>> dist = eq_min_dist(2, N)
    >>> np.round(calc_dist_coeff(2, N, dist), 4)
    array([2.8284, 2.4495, 2.8284, 3.1623, 3.4641])
    """
    return min_euclidean_dist * np.power(N, 1 / dim)


def calc_energy_coeff(dim, N, s, energy):
    r"""
    Coefficient of second term in expansion of energy.

    Parameters
    ----------
    dim : int
        Number of dimensions.
    N : int or array-like
        Number of regions.
    s : float
        Exponent parameter.
    energy : array-like
        Energy values, same shape as N.

    Returns
    -------
    coeff : array-like
        Coefficient(s), same shape as N.

    Notes
    -----
    The returned coefficient `coeff` (denoted $C$) corresponds to the second term
    in the energy expansion. In the PhD thesis :ref:`[Leo07] <leo07>`,
    the "energy coefficient" $ec_d(\mathcal{N})$ is defined as:
    $ec_d(\mathcal{N}) := (1 - E/I(d,s)) \mathcal{N}^{s/d}$.
    For $s = dim-1$ (where $I(d,s)=1$ on $S^2$ and higher), this relates to $C$ as:
    $ec_d(\mathcal{N}) = -2 \times C$.

    The energy expansion is not valid for N == 1,
    and in particular, eq_energy_coeff(dim, N, 0, energy) := 0.

    For s > 0, :ref:`[Kui98] <kui98>` has
    E(dim, N, s) == (sphere_int_energy(dim, s)/2) N^2 + COEFF N^(1+s/dim) + ...

    For s == 0 (logarithmic potential), see :ref:`[Saf97] <saf97>`.

    See Also
    --------
    eq_energy_dist, eq_energy_coeff

    Examples
    --------
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> N = np.arange(2, 7)
    >>> energy, dist = eq_energy_dist(2, N, 0)
    >>> calc_energy_coeff(2, N, 0, energy)
    array([-0.2213, -0.1569, -0.2213, -0.2493, -0.2569])
    """

    def sphere_int_energy_inner(dim, s_val):
        if s_val != 0:
            return (
                math.gamma((dim + 1) / 2)
                * math.gamma(dim - s_val)
                / (math.gamma((dim - s_val + 1) / 2) * math.gamma(dim - s_val / 2))
            )
        if dim != 1:
            return (psi(dim) - psi(dim / 2) - math.log(4)) / 2
        return 0

    N = np.asarray(N)
    energy = np.asarray(energy)
    if s > 0:
        first_term = (sphere_int_energy_inner(dim, s) / 2) * np.power(N, 2)
        coeff = (energy - first_term) / np.power(N, 1 + s / dim)
    else:
        shape = N.shape
        n_partitions = int(np.prod(shape))
        N_flat = N.reshape(1, n_partitions)
        first_term = (sphere_int_energy_inner(dim, s) / 2) * np.power(N_flat, 2)
        coeff = np.zeros_like(N_flat, dtype=float)
        neq1 = N_flat != 1
        coeff[neq1] = (
            energy.reshape(1, n_partitions)[0][neq1.ravel()]
            - first_term[0][neq1.ravel()]
        ) / (N_flat[0][neq1.ravel()] * np.log(N_flat[0][neq1.ravel()]))
        coeff = coeff.reshape(shape)
    return coeff


def sphere_int_energy(dim, s):
    """
    Energy integral of r^(-s) potential.

    Parameters
    ----------
    dim : int
        Number of dimensions.
    s : float
        Exponent parameter.

    Returns
    -------
    energy : float
        Energy integral on S^dim of the r^(-s) potential.

    Notes
    -----
    Ref for s > 0: :ref:`[Kui98] <kui98>`
    Ref for s == 0 and dim == 2: :ref:`[Saf97] <saf97>`

    See Also
    --------
    eq_energy_dist, calc_energy_coeff

    Examples
    --------
    >>> float(sphere_int_energy(2, 0))
    -0.1931471805599453
    """
    if s != 0:
        return (
            math.gamma((dim + 1) / 2)
            * math.gamma(dim - s)
            / (math.gamma((dim - s + 1) / 2) * math.gamma(dim - s / 2))
        )
    if dim != 1:
        return (psi(dim) - psi(dim / 2) - math.log(4)) / 2
    return 0


def calc_packing_density(dim, N, min_euclidean_dist):
    """
    Density of packing given by minimum distance.

    Parameters
    ----------
    dim : int
        Number of dimensions.
    N : int or array-like
        Number of regions.
    min_euclidean_dist : array-like
        Minimum Euclidean distance(s), same shape as N.

    Returns
    -------
    density : array-like
        Density values, same shape as N.

    Notes
    -----
    The packing density is defined as the sum of the areas of the spherical
    caps divided by the area of the unit sphere $S^{dim}$.

    The spherical radius of the caps is half the minimum spherical distance.
    For N == 1, the spherical radius is pi.

    See Also
    --------
    eq_min_dist, area_of_cap, area_of_sphere, eq_packing_density

    Examples
    --------
    >>> np.set_printoptions(precision=4, suppress=True)
    >>> N = np.arange(2, 7)
    >>> dist = eq_min_dist(2, N)
    >>> calc_packing_density(2, N, dist)
    array([1.    , 0.4393, 0.5858, 0.7322, 0.8787])
    """
    s_cap = euc2sph_dist(min_euclidean_dist) / 2
    s_cap = np.array(s_cap)
    N = np.asarray(N)
    s_cap[N == 1] = np.pi
    density = N * area_of_cap(dim, s_cap) / area_of_sphere(dim)
    return density


def eq_dist_coeff(dim, N, extra_offset=False, show_progress=False, even_collars=False):
    """
    Coefficient of minimum distance of an EQ point set.

    Parameters
    ----------
    dim : int
        Number of dimensions.
    N : int or array-like
        Number of regions.
    extra_offset : bool, optional
        Use extra offsets (experimental feature from the original MATLAB
        toolbox for dim 2-3).
        Default False.
    show_progress : bool, optional
        Show progress messages. Default False.
    even_collars : bool, optional
        Use even number of collars for symmetric partitions. Default False.

    Returns
    -------
    coeff : array-like
        Coefficient(s), same shape as N.

    Examples
    --------
    >>> np.round(eq_dist_coeff(2, np.arange(2, 5)), 4)
    array([2.8284, 2.4495, 2.8284])
    >>> float(np.round(eq_dist_coeff(2, 6, show_progress=True), 4))
    3.4641
    """
    dist = eq_min_dist(
        dim,
        N,
        extra_offset=extra_offset,
        show_progress=show_progress,
        even_collars=even_collars,
    )
    coeff = dist * np.power(N, 1 / dim)
    return coeff


def eq_energy_coeff(
    dim, N, s=None, extra_offset=False, show_progress=False, even_collars=False
):
    r"""
    Coefficient in expansion of energy of an EQ point set.

    Parameters
    ----------
    dim : int
        Number of dimensions.
    N : int or array-like
        Number of regions.
    s : float, optional
        Exponent parameter. Defaults to dim-1.
    extra_offset : bool, optional
        Use extra offsets (experimental feature from the original MATLAB
        toolbox for dim 2-3).
        Default False.
    show_progress : bool, optional
        Show progress messages. Default False.
    even_collars : bool, optional
        Use even number of collars for symmetric partitions. Default False.

    Returns
    -------
    coeff : array-like
        Coefficient(s), same shape as N.

    Notes
    -----
    The returned `coeff` ($C$) relates to the thesis metric $ec_d(\mathcal{N})$ as:
    $ec_d(\mathcal{N}) = -2 \times C$ (for $s=dim-1$).
    See Remark on page 198 of :ref:`[Leo07] <leo07>`.

    Examples
    --------
    >>> float(np.round(eq_energy_coeff(2, 4), 4))
    -0.5214
    >>> float(np.round(eq_energy_coeff(2, 6, show_progress=True), 4))
    -0.5453
    """
    if s is None:
        s = dim - 1
    dist_result = eq_energy_dist(
        dim,
        N,
        s=s,
        extra_offset=extra_offset,
        show_progress=show_progress,
        even_collars=even_collars,
    )
    if isinstance(dist_result, tuple):
        energy = dist_result[0]
    else:
        energy = dist_result
    coeff = calc_energy_coeff(dim, N, s, energy)
    return coeff


def eq_energy_dist(
    dim, N, s=None, extra_offset=False, show_progress=False, even_collars=False
):
    """
    Energy and minimum distance of an EQ point set.

    Parameters
    ----------
    dim : int
        Number of dimensions.
    N : int or array-like
        Number of regions.
    s : float, optional
        Exponent parameter. Defaults to dim-1.
    extra_offset : bool, optional
        Use extra offsets (experimental feature from the original MATLAB
        toolbox for dim 2-3).
        Default False.
    show_progress : bool, optional
        Show progress messages. Default False.
    even_collars : bool, optional
        Use even number of collars for symmetric partitions. Default False.

    Returns
    -------
    energy : array-like
        Energy values, same shape as N.
    dist : array-like, optional
        Minimum Euclidean distance(s), same shape as N.

    Examples
    --------
    >>> energy = eq_energy_dist(2, 6, show_progress=True)
    >>> float(np.round(energy, 4))
    9.9853
    """
    if s is None:
        s = dim - 1

    shape = np.shape(N)
    n_partitions = int(np.prod(shape))
    N_flat = np.reshape(N, (1, n_partitions))
    energy = np.zeros_like(N_flat, dtype=float)
    dist = np.zeros_like(N_flat, dtype=float)
    for i, n_val in enumerate(N_flat[0]):
        if show_progress and n_partitions > 1:
            print(f"    N={n_val:6} ({i + 1}/{n_partitions})", end="\r", flush=True)
        points = eq_point_set(dim, n_val, extra_offset, even_collars=even_collars)
        if len(dist.shape) > 1:
            energy[0, i], dist[0, i] = point_set_energy_dist(points, s)
        else:
            energy[0, i] = point_set_energy_dist(points, s)  # pragma: no cover
    if show_progress and n_partitions > 1:
        print()  # Clear the line
    energy = energy.reshape(shape)
    dist = dist.reshape(shape)
    if len(dist.shape) > 0:
        return energy, dist
    return energy


def eq_min_dist(dim, N, extra_offset=False, show_progress=False, even_collars=False):
    """
    Minimum distance between centre points of an EQ partition.

    Parameters
    ----------
    dim : int
        Number of dimensions.
    N : int or array-like
        Number of regions.
    extra_offset : bool, optional
        Use extra offsets (experimental feature from the original MATLAB
        toolbox for dim 2-3).
        Default False.
    show_progress : bool, optional
        Show progress messages. Default False.
    even_collars : bool, optional
        Use even number of collars for symmetric partitions. Default False.

    Returns
    -------
    dist : array-like
        Minimum Euclidean distance(s), same shape as N.

    Notes
    -----
    Exploits the collar structure for efficient calculation.

    Examples
    --------
    >>> float(np.round(eq_min_dist(2, 6, show_progress=True), 4))
    1.4142
    """
    shape = np.shape(N)
    n_partitions = int(np.prod(shape))
    N_flat = np.reshape(N, (1, n_partitions))
    dist = np.zeros_like(N_flat, dtype=float)
    for i, n_val in enumerate(N_flat[0]):
        if show_progress and n_partitions > 1:
            print(f"    N={n_val:6} ({i + 1}/{n_partitions})", end="\r", flush=True)
        dist[0, i] = _eq_min_dist_scalar(
            dim, int(n_val), extra_offset, even_collars=even_collars
        )
    if show_progress and n_partitions > 1:
        print()  # Clear the line
    return dist.reshape(shape)


def _eq_min_dist_scalar(dim, N, extra_offset=False, even_collars=False):
    """
    Scalar version of eq_min_dist.
    """
    if N <= 1:
        return 2.0
    if dim == 1:
        # Distance on a circle with N points: 2 * sin(pi/N)
        return 2 * np.sin(np.pi / N)

    _, n_regions = eq_caps(dim, N, even_collars=even_collars)

    # Exploiting the collar structure:
    # 1. Intra-collar distances
    # 2. Inter-collar distances (adjacent only)
    #
    # We localized the KDTree search to adjacent collars to achieve
    # near-linear scaling and low memory usage.

    point_sets = []
    points = eq_point_set(dim, N, extra_offset, even_collars=even_collars)
    idx = 0
    for n_k in n_regions:
        nk = int(n_k)
        point_sets.append(points[:, idx : idx + nk])
        idx += nk

    d_min = 2.0
    for k, n_k in enumerate(n_regions):
        # Intra-collar
        if n_k > 1:
            d_intra = point_set_min_dist(point_sets[k])
            d_min = min(d_min, d_intra)
        # Inter-collar with NEXT
        if k < len(n_regions) - 1:
            # Query the KDTree of collar k with the points of collar k+1
            tree = KDTree(point_sets[k].T)
            dists, _ = tree.query(point_sets[k + 1].T, k=1)
            d_inter = np.min(dists)
            d_min = min(d_min, d_inter)

    return d_min


def eq_packing_density(
    dim, N, extra_offset=False, show_progress=False, even_collars=False
):
    """
    Density of packing given by minimum distance of EQ point set.

    Parameters
    ----------
    dim : int
        Number of dimensions.
    N : int or array-like
        Number of regions.
    extra_offset : bool, optional
        Use extra offsets (experimental feature from the original MATLAB
        toolbox for dim 2-3).
        Default False.
    show_progress : bool, optional
        Show progress messages. Default False.
    even_collars : bool, optional
        Use even number of collars for symmetric partitions. Default False.

    Returns
    -------
    density : array-like
        Density values, same shape as N.

    Examples
    --------
    >>> float(np.round(eq_packing_density(2, 4), 4))
    0.5858
    >>> float(np.round(eq_packing_density(2, 6, show_progress=True), 4))
    0.8787
    """
    min_euclidean_dist = eq_min_dist(
        dim,
        N,
        extra_offset=extra_offset,
        show_progress=show_progress,
        even_collars=even_collars,
    )
    density = calc_packing_density(dim, N, min_euclidean_dist)
    return density


def eq_point_set_property(fhandle, dim, N, extra_offset=False, show_progress=False):
    """
    Property of an EQ point set.

    Parameters
    ----------
    fhandle : callable
        Function expecting an array (dim+1 x N), returns property value.
    dim : int
        Number of dimensions.
    N : int or array-like
        Number of regions.
    extra_offset : bool, optional
        Use extra offsets (experimental feature from the original MATLAB
        toolbox for dim 2-3).
        Default False.
    show_progress : bool, optional
        Show progress messages. Default False.

    Returns
    -------
    property : array-like
        Property value(s), same shape as N.

    Examples
    --------
    >>> import numpy as np
    >>> f = lambda x: np.mean(x)
    >>> float(np.round(eq_point_set_property(f, 2, 6, show_progress=True), 10))
    0.0
    """
    shape = np.shape(N)
    n_partitions = int(np.prod(shape))
    N_flat = np.reshape(N, (1, n_partitions))
    property_vals = np.zeros_like(N_flat, dtype=float)
    for i, n_val in enumerate(N_flat[0]):
        if show_progress and n_partitions > 1:
            print(f"    N={n_val:6} ({i + 1}/{n_partitions})", end="\r", flush=True)
        points = eq_point_set(dim, n_val, extra_offset)
        property_vals[0, i] = fhandle(points)
    if show_progress and n_partitions > 1:
        print()  # Clear the line
    property_vals = property_vals.reshape(shape)
    return property_vals


def point_set_dist_coeff(points):
    """
    Coefficient of minimum distance of a point set.

    Parameters
    ----------
    points : array-like
        Array of shape (dim+1, N), columns are points in R^{dim+1}.

    Returns
    -------
    coeff : float
        Coefficient value.

    Notes
    -----
    For details, see calc_dist_coeff.

    See Also
    --------
    point_set_min_dist, calc_dist_coeff, eq_dist_coeff, eq_min_dist

    Examples
    --------
    >>> x = np.array([[0, 0, 0, 0], [0, 1, -1, 0], [1, 0, 0, -1]])
    >>> float(point_set_dist_coeff(x))
    2.8284271247461903
    """
    dim = points.shape[0] - 1
    N = points.shape[1]
    min_euclidean_dist = point_set_min_dist(points)
    coeff = calc_dist_coeff(dim, N, min_euclidean_dist)
    return coeff


def point_set_energy_coeff(points, s=None):
    """
    Coefficient in expansion of energy of a point set.

    Parameters
    ----------
    points : array-like
        Array of shape (dim+1, N), columns are points in R^{dim+1}.
    s : float, optional
        Exponent parameter. Defaults to dim-1.

    Returns
    -------
    coeff : float
        Coefficient value.

    Notes
    -----
    For details, see calc_energy_coeff.

    See Also
    --------
    point_set_energy_dist, calc_energy_coeff, eq_energy_coeff, eq_energy_dist

    Examples
    --------
    >>> x = np.array([[0, 0, 0, 0], [0, 1, -1, 0], [1, 0, 0, -1]])
    >>> point_set_energy_coeff(x)
    array([-0.5214, -0.8232])
    """
    dim = points.shape[0] - 1
    N = points.shape[1]
    if s is None:
        s = dim - 1
    energy = point_set_energy_dist(points, s)
    coeff = calc_energy_coeff(dim, N, s, energy)
    return coeff


def point_set_energy_dist(points, s=None, block_size=2000):
    """
    Energy and minimum distance of a point set.

    Parameters
    ----------
    points : array-like
        Array of shape (M, N), columns are points in R^M.
    s : float, optional
        Exponent parameter. Defaults to dim-1.
    block_size : int, optional
        Maximum number of points to process in a single block.
        Defaults to 2000, balancing memory usage and vectorization overhead.

    Returns
    -------
    energy : float
        Energy value.
    min_dist : float
        Minimum Euclidean distance.

    Notes
    -----
    ENERGY for single point is 0. MIN_DIST for single point is 2.

    See Also
    --------
    eq_energy_dist, point_set_min_dist

    Examples
    --------
    >>> x = np.array([[0, 0, 0, 0], [0, 1, -1, 0], [1, 0, 0, -1]])
    >>> energy, dist = point_set_energy_dist(x, 2)
    >>> (float(energy), float(dist))
    (2.5, 1.4142135623730951)
    >>> energy, dist = point_set_energy_dist(x, 0)
    >>> (float(energy), float(dist))
    (-2.7725887222397816, 1.4142135623730951)
    """
    M, N = points.shape
    dim = M - 1
    if s is None:
        s = dim - 1

    # Handle N=1 case
    if N <= 1:
        return 0.0, 2.0

    energy = 0.0
    min_dist = np.inf

    # Process in blocks to limit peak memory usage to O(block_size^2)
    # and exploit symmetry (only compute upper triangle of block matrix)
    for i in range(0, N, block_size):
        end_i = min(i + block_size, N)
        pts_i = points[:, i:end_i].T

        for j in range(i, N, block_size):
            end_j = min(j + block_size, N)
            pts_j = points[:, j:end_j].T

            # Efficient pairwise distance between block i and block j
            dists = cdist(pts_i, pts_j, metric="euclidean")

            if i == j:
                # Diagonal block: mask self-distances
                np.fill_diagonal(dists, np.inf)

                # Only take upper triangle to avoid double counting within the block
                valid_mask = np.triu(np.ones_like(dists, dtype=bool), k=1)
                valid_dists = dists[valid_mask]

                if valid_dists.size > 0:
                    min_dist = min(min_dist, np.min(valid_dists))
            else:
                # Off-diagonal block: all distances are valid and distinct
                # We calculate i<j, representing half the symmetric interaction
                valid_dists = dists.flatten()
                min_dist = min(min_dist, np.min(valid_dists))

            # Accumulate energy contribution
            if valid_dists.size > 0:
                if s == 0:
                    energy += np.sum(-np.log(valid_dists))
                else:
                    energy += np.sum(np.power(valid_dists, -s))

    return energy, min_dist


def point_set_min_dist(points):
    """
    Minimum distance between points of a point set.

    Parameters
    ----------
    points : array-like
        Array of shape (dim+1, N).

    Returns
    -------
    min_dist : float
        Minimum Euclidean distance.

    Notes
    -----
    Uses scipy.spatial.KDTree for efficient O(N log N) calculation.
    """
    _, N = points.shape
    if N <= 1:
        return 2.0
    tree = KDTree(points.T)
    # query(k=2) returns the distance to itself (0) and the nearest neighbor
    dists, _ = tree.query(points.T, k=2)
    min_dist = np.min(dists[:, 1])
    return min_dist


if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod()



"""
PyEQSP partitions module.

Copyright 2026 Paul Leopardi
"""

import math
from math import pi

import numpy as np

from ._private._partitions import (
    bot_cap_region,
    cap_colats,
    centres_of_regions,
    circle_offset,
    ideal_region_list,
    num_collars,
    polar_colat,
    round_to_naturals,
    s2_offset,
    sphere_region,
    top_cap_region,
)
from .utilities import (
    asfloat,
    cart2polar2,
    ideal_collar_angle,
    polar2cart,
)

TAU = 2.0 * pi


def eq_caps(dim, N, even_collars=False):
    """
    Partition a sphere into two nested spherical caps.

    Parameters
    ----------
    dim : int
        The number of dimensions.
    N : int
        The number of regions.
    even_collars : bool, optional
        If True, force an even number of collars so the equatorial
        hyperplane falls exactly on a cap boundary. This enables
        clean hemisphere splitting for symmetric sampling methods.
        Requires N to be an even number.
        Default is False.

    Returns
    -------
    s_cap : ndarray
        1D array containing increasing colatitudes of caps. Size is
        (n_collars+2,).
    n_regions : ndarray
        1D array containing the integer number of regions in each zone.
        Size is (n_collars+2,).

    Raises
    ------
    ValueError
        If dim or N are not positive integers, or if even_collars is True
        but N is odd.

    See Also
    --------
    eq_regions, eq_point_set_polar

    Notes
    -----
    - s_cap[0] is the colatitude of the North polar cap.
    - s_cap[-2] is pi - c_polar.
    - s_cap[-1] is pi.
    - n_regions[0] is 1 and n_regions[-1] is 1.
    - The sum of n_regions equals N.

    Examples
    --------
    >>> import numpy as np
    >>> np.set_printoptions(precision=4)
    >>> s_cap, n_regions = eq_caps(2,10)
    >>> s_cap
    array([0.6435, 1.5708, 2.4981, 3.1416])
    >>> n_regions
    array([1., 4., 4., 1.])

    >>> s_cap, n_regions = eq_caps(3,6)
    >>> s_cap
    array([0.9845, 2.1571, 3.1416])
    >>> n_regions
    array([1., 4., 1.])

    >>> s_cap, n_regions = eq_caps(2, 10, even_collars=True)
    >>> s_cap
    array([0.6435, 1.5708, 2.4981, 3.1416])
    >>> n_regions
    array([1., 4., 4., 1.])
    """
    if not (isinstance(dim, (int, np.integer)) and dim >= 1):
        raise ValueError("dim must be a positive integer")
    if not (isinstance(N, (int, np.integer)) and N >= 1):
        raise ValueError("N must be a positive integer")

    if dim == 1:
        # Circle: return the angles of N equal sectors.
        sector = np.arange(1, N + 1)
        s_cap = sector * TAU / N
        n_regions = np.ones_like(sector, dtype=int)
    elif N == 1:
        # Only one region: whole sphere.
        s_cap = np.array([pi])
        n_regions = np.array([1], dtype=int)
    else:
        # Determine polar colatitude
        c_polar = polar_colat(dim, N)
        a_ideal = ideal_collar_angle(dim, N)

        if even_collars and N % 2 != 0:
            raise ValueError("even_collars=True requires N to be an even number")

        if even_collars and N > 2:
            # Force even number of collars so equator falls on a cap boundary
            ratio_half = 0.5 * (pi - 2.0 * c_polar) / a_ideal
            n_half = max(1, round(ratio_half))
            n_collars = 2 * n_half
        else:
            # Determine number of collars
            n_collars = num_collars(N, c_polar, a_ideal)

        # Ideal real number of regions per collar
        r_regions = ideal_region_list(dim, N, c_polar, n_collars)
        # Round to natural numbers preserving sum N
        n_regions = round_to_naturals(N, r_regions)
        # Colatitudes of cap tops
        s_cap = cap_colats(dim, N, c_polar, n_regions)

    return asfloat(s_cap), asfloat(n_regions)


def eq_point_set(dim, N, extra_offset=False, even_collars=False):
    """
    Centre points of regions of EQ partition, in Cartesian coordinates.

    Parameters
    ----------
    dim : int
        The number of dimensions.
    N : int
        The number of regions.
    extra_offset : bool, optional
        If True, enables experimental extra offsets for dim 2 and 3. This is
        a feature from the original MATLAB toolbox and is no longer planned
        for expansion to dim > 3.
    even_collars : bool, optional
        If True, force an even number of collars. See `eq_caps` for details.

    Returns
    -------
    points_x : ndarray
        Array of shape (dim+1, N) containing centre points of each region in
        Cartesian coordinates. Each column is a point on S^dim.

    Raises
    ------
    ValueError
        If dim or N are not positive integers.

    See Also
    --------
    eq_point_set_polar, polar2cart, eq_regions

    Notes
    -----
    Uses the recursive zonal equal area algorithm. The optional argument
    "offset", "extra" may be provided and is forwarded to
    eq_point_set_polar.

    Examples
    --------
    >>> points = eq_point_set(2, 4)  # doctest: +ELLIPSIS
    >>> points.shape
    (3, 4)

    >>> points = eq_point_set(2, 4, even_collars=True)
    >>> points.shape
    (3, 4)
    """
    points_polar = eq_point_set_polar(
        dim, N, extra_offset=extra_offset, even_collars=even_collars
    )
    points_x = polar2cart(points_polar)
    return points_x


def eq_point_set_polar(dim, N, extra_offset=False, even_collars=False):
    """
    Centre points of regions of an EQ partition in polar coordinates.

    Parameters
    ----------
    dim : int
        The spatial dimension of the sphere (S^dim in R^{dim+1}).
    extra_offset : bool, optional
        If True, enables experimental extra offsets for dim 2 and 3. This is
        a feature from the original MATLAB toolbox and is no longer planned
        for expansion to dim > 3.
    even_collars : bool, optional
        If True, force an even number of collars. See `eq_caps` for details.

    Returns
    -------
    points_s : ndarray
        Array of shape (dim, N) containing centre points in spherical polar
        coordinates. Each column is a point on S^dim.

    Raises
    ------
    ValueError
        If dim or N are not positive integers.

    See Also
    --------
    eq_point_set, polar2cart, cart2polar2

    Notes
    -----
    - If dim > 3, extra offsets are ignored.
    - Points are arranged such that each region centre is the centre of the
      corresponding product interval in spherical coordinates, except for
      polar caps where the spherical cap centre is used.

    Examples
    --------
    >>> pts = eq_point_set_polar(2, 4)  # doctest: +ELLIPSIS
    >>> pts.shape
    (2, 4)

    >>> pts = eq_point_set_polar(2, 4, even_collars=True)
    >>> pts.shape
    (2, 4)
    """
    # Extra offsets not used for dim > 3
    if dim > 3:
        extra_offset = False

    if not (isinstance(dim, (int, np.integer)) and dim >= 1):
        raise ValueError("dim must be a positive integer")
    if not (isinstance(N, (int, np.integer)) and N >= 1):
        raise ValueError("N must be a positive integer")

    if N == 1:
        points_s = np.zeros((dim, 1))
        return points_s

    a_cap, n_regions = eq_caps(dim, N, even_collars=even_collars)
    # a_cap is increasing list of angles of caps

    if dim == 1:
        # Circle: points placed half way along each sector
        points_s = a_cap - pi / N
        points_s = np.asarray(points_s).reshape((1, N))
        return points_s

    n_collars = int(np.size(n_regions) - 2)
    use_cache = dim >= 2
    cache = None
    if use_cache:
        cache_size = n_collars // 2
        cache = [None] * max(1, cache_size)

    points_s = np.zeros((dim, N))
    point_n = 1  # MATLAB indexing: points_s column 1 is north pole
    # North polar cap 'centre' is North pole (all zeros except last angle 0)
    # Represented by zeros column; will fill first column later
    # For extra_offset rotation bookkeeping
    if extra_offset and (dim == 3):
        R = np.eye(3)
    if dim == 2:
        offset = 0.0

    # Start with north pole
    points_s[:, 0] = 0.0
    point_n = 1

    for collar_n in range(1, n_collars + 1):
        a_top = a_cap[collar_n - 1]
        a_bot = a_cap[collar_n]
        n_in_collar = int(n_regions[collar_n])

        # Partition the (dim-1)-sphere into n_in_collar regions
        if use_cache:
            twin_collar_n = n_collars - collar_n + 1
            if (
                twin_collar_n <= (len(cache))
                and cache[twin_collar_n - 1] is not None
                and cache[twin_collar_n - 1].shape[1] == n_in_collar
            ):
                points_1 = cache[twin_collar_n - 1]
            else:
                points_1 = eq_point_set_polar(dim - 1, n_in_collar, extra_offset)
                if collar_n <= len(cache):
                    cache[collar_n - 1] = points_1
        else:
            points_1 = eq_point_set_polar(
                dim - 1, n_in_collar, extra_offset
            )  # pragma: no cover

        if extra_offset and (dim == 3) and (collar_n > 1):
            # Rotate 2-spheres to prevent alignment of north poles.
            R = s2_offset(points_1) @ R  # pragma: no cover
            points_1 = cart2polar2(R @ polar2cart(points_1))  # pragma: no cover

        a_point = (a_top + a_bot) / 2.0
        n_points_1 = points_1.shape[1]
        point_idx = np.arange(point_n, point_n + n_points_1)

        if dim == 2:
            # 1D angles on circle
            pts = points_1[:, :].flatten()
            pts = np.mod(pts + TAU * offset, TAU)
            points_s[0, point_idx] = pts
            # Update offset
            next_n = (
                int(n_regions[collar_n + 1]) if (collar_n + 1) < len(n_regions) else 0
            )
            offset = offset + circle_offset(n_in_collar, next_n, extra_offset)
            offset = offset - math.floor(offset)
        else:
            points_s[0 : dim - 1, point_idx] = points_1[:, :]

        points_s[dim - 1, point_idx] = a_point
        point_n = point_n + n_points_1

    # Bottom polar cap centre
    points_s[:, point_n] = 0.0
    points_s[dim - 1, point_n] = pi
    return points_s


def eq_regions(dim, N, extra_offset=False, even_collars=False):
    """
    Recursive zonal equal area (EQ) partition of sphere.

    Parameters
    ----------
    dim : int
        The spatial dimension of the sphere (S^dim).
    extra_offset : bool, optional
        If True, enables experimental extra offsets for dim 2 and 3. This is
        a feature from the original MATLAB toolbox and is no longer planned
        for expansion to dim > 3.
    even_collars : bool, optional
        If True, force an even number of collars. See `eq_caps` for details.

    Returns
    -------
    regions : ndarray
        An array of shape (dim, 2, N) containing the boundaries of the N regions
        in spherical polar coordinates.
        For each region r in 0..N-1:
        - regions[i, 0, r] is the lower bound of the i-th coordinate,
        - regions[i, 1, r] is the upper bound of the i-th coordinate.
    dim_1_rot : list (optional)
        If requested (by caller), a list of N rotation matrices, each of size
        (dim, dim), describing R^dim rotations needed to place regions when
        extra offsets are used (only meaningful for dim == 3).

    Raises
    ------
    ValueError
        If dim or N are not positive integers.

    See Also
    --------
    eq_point_set, centres_of_regions

    Notes
    -----
    - For N == 1, the single region is the whole sphere.
    - If extra_offset is used and dim == 3, the returned dim_1_rot describes
      the rotation applied to sub-spheres.

    Examples
    --------
    >>> regs = eq_regions(2, 4)  # doctest: +ELLIPSIS
    >>> regs.shape
    (2, 2, 4)

    >>> regs = eq_regions(2, 4, even_collars=True)
    >>> regs.shape
    (2, 2, 4)
    """

    if dim > 3:
        extra_offset = False

    if not (isinstance(dim, (int, np.integer)) and dim >= 1):
        raise ValueError("dim must be a positive integer")
    if not (isinstance(N, (int, np.integer)) and N >= 1):
        raise ValueError("N must be a positive integer")

    dim_1_rot = None
    # Prepare output rotation containers if caller wants them.
    dim_1_rot = [None] * N

    if N == 1:
        regions = np.zeros((dim, 2, 1))
        regions[:, :, 0] = sphere_region(dim)
        if extra_offset and dim == 3:
            dim_1_rot[0] = np.eye(dim)
            return regions, dim_1_rot
        return regions

    s_cap, n_regions = eq_caps(dim, N, even_collars=even_collars)

    if dim == 1:
        # Circle: return pairs of sector angles
        regions = np.zeros((dim, 2, N))
        if N > 1:
            regions[:, 0, 1:N] = s_cap[0 : N - 1]
        regions[:, 1, :] = s_cap
        # rotations are identity
        if extra_offset:
            for idx in range(N):
                dim_1_rot[idx] = np.eye(dim)
            return regions, dim_1_rot
        return regions

    n_collars = int(np.size(n_regions) - 2)
    use_cache = dim > 2
    cache = None
    if use_cache:
        cache_size = n_collars // 2
        cache = [None] * max(1, cache_size)

    regions = np.zeros((dim, 2, N))
    # Top cap
    regions[:, :, 0] = top_cap_region(dim, s_cap[0])
    region_n = 0

    R = np.eye(dim)
    dim_1_rot[0] = R

    if dim == 2:
        offset = 0.0

    for collar_n in range(1, n_collars + 1):
        c_top = s_cap[collar_n - 1]
        c_bot = s_cap[collar_n]
        n_in_collar = int(n_regions[collar_n])

        # Partition the (dim-1)-sphere
        if use_cache:
            twin_collar_n = n_collars - collar_n + 1
            if (
                twin_collar_n <= len(cache)
                and cache[twin_collar_n - 1] is not None
                and cache[twin_collar_n - 1].shape[2] == n_in_collar
            ):
                regions_1 = cache[twin_collar_n - 1]
            else:
                if extra_offset:
                    regions_1, _ = eq_regions(dim - 1, n_in_collar, extra_offset)
                else:
                    regions_1 = eq_regions(dim - 1, n_in_collar, extra_offset)
                if collar_n <= len(cache):
                    cache[collar_n - 1] = regions_1
        else:
            if extra_offset:
                regions_1, _ = eq_regions(dim - 1, n_in_collar, extra_offset)
            else:
                regions_1 = eq_regions(dim - 1, n_in_collar, extra_offset)

        if extra_offset and (dim == 3) and (collar_n > 1):
            R = s2_offset(centres_of_regions(regions_1)) @ R

        # Append regions for this collar
        n_regions_1 = regions_1.shape[2]
        idx_start = region_n + 1
        idx_end = region_n + 1 + n_regions_1

        if dim == 2:
            r_top_0 = np.mod(regions_1[0, 0, :] + TAU * offset, TAU)
            r_bot_0 = np.mod(regions_1[0, 1, :] + TAU * offset, TAU)

            mask = r_bot_0 < r_top_0
            r_bot_0[mask] += TAU

            regions[0, 0, idx_start:idx_end] = r_top_0
            regions[0, 1, idx_start:idx_end] = r_bot_0
            regions[1, 0, idx_start:idx_end] = c_top
            regions[1, 1, idx_start:idx_end] = c_bot
        else:
            regions[0 : dim - 1, :, idx_start:idx_end] = regions_1
            regions[dim - 1, 0, idx_start:idx_end] = c_top
            regions[dim - 1, 1, idx_start:idx_end] = c_bot

        if dim_1_rot is not None:
            dim_1_rot[idx_start:idx_end] = [R] * n_regions_1

        region_n += n_regions_1

        if dim == 2:
            next_n = (
                int(n_regions[collar_n + 1]) if (collar_n + 1) < len(n_regions) else 0
            )
            offset = offset + circle_offset(n_in_collar, next_n, extra_offset)
            offset = offset - math.floor(offset)

    # Bottom cap
    regions[:, :, N - 1] = bot_cap_region(dim, s_cap[0])
    if extra_offset:
        if (dim == 3) or (dim < 3):
            # Ensure last one is set if not already
            if dim_1_rot[N - 1] is None:
                dim_1_rot[N - 1] = np.eye(dim)
        return regions, dim_1_rot
    return regions
